# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Handle Ansible collections.
"""

from __future__ import annotations

import json
import os
import threading
import typing as t
from collections.abc import Collection, Iterator, Sequence
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path

from antsibull_fileutils.yaml import load_yaml_file

from ..ansible import AnsibleCoreVersion
from .data import CollectionData

# Function that runs a command (and fails on non-zero return code)
# and returns a tuple (stdout, stderr)
Runner = t.Callable[[list[str]], tuple[bytes, bytes]]


GALAXY_YML = "galaxy.yml"
MANIFEST_JSON = "MANIFEST.json"


@dataclass(frozen=True)
class _GlobalCache:
    root: Path
    download_cache: Path
    extracted_cache: Path

    @classmethod
    def create(cls, *, root: Path) -> _GlobalCache:
        """
        Create a global cache object.
        """
        return cls(
            root=root,
            download_cache=root / "downloaded",
            extracted_cache=root / "extracted",
        )

    def get_extracted_path(self, *, ansible_core_version: AnsibleCoreVersion) -> Path:
        """
        Given an ansible-core version, returns its extracted collection cache directory.
        """
        return self.extracted_cache / str(ansible_core_version)


def _load_galaxy_yml(galaxy_yml: Path) -> dict[str, t.Any]:
    try:
        data = load_yaml_file(galaxy_yml)
    except Exception as exc:
        raise ValueError(f"Cannot parse {galaxy_yml}: {exc}") from exc
    if not isinstance(data, dict):
        raise ValueError(f"{galaxy_yml} is not a dictionary")
    return data


def _load_manifest_json_collection_info(manifest_json: Path) -> dict[str, t.Any]:
    try:
        with open(manifest_json, "br") as f:
            data = json.load(f)
    except Exception as exc:
        raise ValueError(f"Cannot parse {manifest_json}: {exc}") from exc
    ci = data.get("collection_info")
    if not isinstance(ci, dict):
        raise ValueError(f"{manifest_json} does not contain collection_info")
    return ci


def load_collection_data_from_disk(
    path: Path,
    *,
    namespace: str | None = None,
    name: str | None = None,
    root: Path | None = None,
    current: bool = False,
    accept_manifest: bool = True,
) -> CollectionData:
    """
    Load collection data from disk.
    """
    galaxy_yml = path / GALAXY_YML
    manifest_json = path / MANIFEST_JSON
    found: Path
    if galaxy_yml.is_file():
        found = galaxy_yml
        data = _load_galaxy_yml(galaxy_yml)
    elif not accept_manifest:
        raise ValueError(f"Cannot find {GALAXY_YML} in {path}")
    elif manifest_json.is_file():
        found = manifest_json
        data = _load_manifest_json_collection_info(manifest_json)
    else:
        raise ValueError(f"Cannot find {GALAXY_YML} or {MANIFEST_JSON} in {path}")

    ns = data.get("namespace")
    if not isinstance(ns, str):
        raise ValueError(f"{found} does not contain a namespace")
    n = data.get("name")
    if not isinstance(n, str):
        raise ValueError(f"{found} does not contain a name")
    v = data.get("version")
    if not isinstance(v, str):
        v = None
    d = data.get("dependencies") or {}
    if not isinstance(d, dict):
        raise ValueError(f"{found}'s dependencies is not a mapping")

    if namespace is not None and ns != namespace:
        raise ValueError(
            f"{found} contains namespace {ns!r}, but was hoping for {namespace!r}"
        )
    if name is not None and n != name:
        raise ValueError(f"{found} contains name {n!r}, but was hoping for {name!r}")
    return CollectionData(
        collections_root_path=root,
        path=path,
        namespace=ns,
        name=n,
        full_name=f"{ns}.{n}",
        version=v,
        dependencies=d,
        current=current,
    )


def _list_adjacent_collections_ansible_collections_tree(
    root: Path,
    *,
    directories_to_ignore: Collection[Path] | None = None,
) -> Iterator[CollectionData]:
    directories_to_ignore = directories_to_ignore or ()
    for namespace in root.iterdir():  # pylint: disable=too-many-nested-blocks
        try:
            if namespace.is_dir() or namespace.is_symlink():
                for name in namespace.iterdir():
                    if name in directories_to_ignore:
                        continue
                    try:
                        if name.is_dir() or name.is_symlink():
                            yield load_collection_data_from_disk(
                                name,
                                namespace=namespace.name,
                                name=name.name,
                                root=root,
                            )
                    except Exception:  # pylint: disable=broad-exception-caught
                        # If name doesn't happen to be a (symlink to a) directory,
                        # is not readable, ...
                        pass
        except Exception:  # pylint: disable=broad-exception-caught
            # If namespace doesn't happen to be a (symlink to a) directory, is not readable, ...
            pass


def _list_adjacent_collections_outside_tree(
    directory: Path,
    *,
    directories_to_ignore: Collection[Path] | None = None,
) -> Iterator[CollectionData]:
    directories_to_ignore = directories_to_ignore or ()
    for collection_dir in directory.iterdir():
        if collection_dir in directories_to_ignore:
            continue
        if not collection_dir.is_dir() and not collection_dir.is_symlink():
            continue
        parts = collection_dir.name.split(".")
        if len(parts) != 2:
            continue
        namespace, name = parts
        if not namespace.isidentifier() or not name.isidentifier():
            continue
        try:
            yield load_collection_data_from_disk(
                collection_dir,
                namespace=namespace,
                name=name,
            )
        except Exception:  # pylint: disable=broad-exception-caught
            # If collection_dir doesn't happen to be a (symlink to a) directory, ...
            pass


def _fs_list_local_collections() -> Iterator[CollectionData]:
    root: Path | None = None

    # Determine potential root
    cwd = Path.cwd()
    parents: Sequence[Path] = cwd.parents
    if len(parents) > 2 and parents[1].name == "ansible_collections":
        root = parents[1]

    # Current collection
    try:
        current = load_collection_data_from_disk(cwd, root=root, current=True)
        if root and current.namespace == parents[0].name and current.name == cwd.name:
            yield current
        else:
            root = None
            current = load_collection_data_from_disk(cwd, current=True)
            yield current
    except Exception as exc:
        raise ValueError(
            f"Cannot load current collection's info from {cwd}: {exc}"
        ) from exc

    # Search tree
    if root:
        yield from _list_adjacent_collections_ansible_collections_tree(
            root, directories_to_ignore=(cwd,)
        )
    elif len(parents) > 0:
        yield from _list_adjacent_collections_outside_tree(
            parents[0], directories_to_ignore=(cwd,)
        )
    else:
        # Only happens if cwd == "/"
        pass  # pragma: no cover


def _fs_list_global_cache(global_cache_dir: Path) -> Iterator[CollectionData]:
    if not global_cache_dir.is_dir():
        return

    yield from _list_adjacent_collections_outside_tree(global_cache_dir)


def _galaxy_list_collections(runner: Runner) -> Iterator[CollectionData]:
    try:
        stdout, _ = runner(["ansible-galaxy", "collection", "list", "--format", "json"])
        data = json.loads(stdout)
        for collections_root_path, collections in data.items():
            root = Path(collections_root_path)
            for collection in collections:
                namespace, name = collection.split(".", 1)
                try:
                    yield load_collection_data_from_disk(
                        root / namespace / name,
                        namespace=namespace,
                        name=name,
                        root=root,
                        current=False,
                    )
                except:  # noqa: E722, pylint: disable=bare-except
                    # Looks like Ansible passed crap on to us...
                    pass
    except Exception as exc:
        raise ValueError(f"Error while loading collection list: {exc}") from exc


@dataclass
class CollectionList:
    """
    A list of Ansible collections.
    """

    collections: list[CollectionData]
    collection_map: dict[str, CollectionData]
    current: CollectionData

    @classmethod
    def create(cls, collections_map: dict[str, CollectionData]):
        """
        Given a dictionary mapping collection names to collection data, creates a CollectionList.

        One of the collections must have the ``current`` flag set.
        """
        collections = sorted(collections_map.values(), key=lambda cli: cli.full_name)
        current = next(c for c in collections if c.current)
        return cls(
            collections=collections,
            collection_map=collections_map,
            current=current,
        )

    @classmethod
    def collect_global(cls, *, runner: Runner) -> CollectionList:
        """
        Search for a global list of collections. The result is not cached.
        """
        found_collections = {}
        for collection_data in _fs_list_local_collections():
            found_collections[collection_data.full_name] = collection_data
        if os.environ.get("ANTSIBULL_NOX_IGNORE_INSTALLED_COLLECTIONS") != "true":
            for collection_data in _galaxy_list_collections(runner):
                # Similar to Ansible, we use the first match
                if collection_data.full_name not in found_collections:
                    found_collections[collection_data.full_name] = collection_data
        return cls.create(found_collections)

    @classmethod
    def collect_local(
        cls,
        *,
        ansible_core_version: AnsibleCoreVersion,
        global_cache: _GlobalCache,
        current: CollectionData,
    ) -> CollectionList:
        """
        Search for a list of collections from a local cache path. The result is not cached.
        """
        found_collections = {
            current.full_name: current,
        }
        for collection_data in _fs_list_global_cache(
            global_cache.get_extracted_path(ansible_core_version=ansible_core_version)
        ):
            # Similar to Ansible, we use the first match
            if collection_data.full_name not in found_collections:
                found_collections[collection_data.full_name] = collection_data
        return cls.create(found_collections)

    def find(self, name: str) -> CollectionData | None:
        """
        Find a collection for a given name.
        """
        return self.collection_map.get(name)

    def clone(self) -> CollectionList:
        """
        Create a clone of this list.
        """
        return CollectionList(
            collections=list(self.collections),
            collection_map=dict(self.collection_map),
            current=self.current,
        )

    def merge_with(self, other: CollectionList) -> CollectionList:
        """
        Merge this collection list with another (local) one.
        """
        result = dict(self.collection_map)
        for collection in other.collections:
            # Similar to Ansible, we use the first match.
            # For merge, this means we prefer self over other.
            if collection.full_name not in result:
                result[collection.full_name] = collection
        return CollectionList.create(result)

    def _add(self, collection: CollectionData, *, force: bool = True) -> bool:
        if not force and collection.full_name in self.collection_map:
            return False
        self.collections.append(collection)
        self.collection_map[collection.full_name] = collection
        return True


class _CollectionListUpdater:
    def __init__(
        self,
        *,
        owner: "_CollectionListSingleton",
        merged_collection_list: CollectionList,
        local_collection_list: CollectionList,
        ansible_core_version: AnsibleCoreVersion,
    ) -> None:
        self._owner = owner
        self._merged_collection_list = merged_collection_list
        self._local_collection_list = local_collection_list
        self._ansible_core_version = ansible_core_version

    def find(self, name: str) -> CollectionData | None:
        """
        Find a collection for a given name.
        """
        return self._merged_collection_list.find(name)

    def add_collection(
        self, *, directory: Path, namespace: str, name: str
    ) -> CollectionData:
        """
        Add a new collection to the cache.
        """
        # pylint: disable-next=protected-access
        result = self._owner._add_collection(
            directory=directory,
            namespace=namespace,
            name=name,
            ansible_core_version=self._ansible_core_version,
        )
        self._merged_collection_list._add(result)  # pylint: disable=protected-access
        self._local_collection_list._add(result)  # pylint: disable=protected-access
        return result

    def get_global_cache(self) -> _GlobalCache:
        """
        Get the global cache object.
        """
        return self._owner._get_global_cache()  # pylint: disable=protected-access


class _CollectionListSingleton:
    _lock = threading.Lock()

    _global_cache_dir: Path | None = None
    _global_collection_list: CollectionList | None = None
    _global_collection_list_per_ansible_core_version: dict[
        AnsibleCoreVersion, CollectionList
    ] = {}

    def setup(self, *, global_cache_dir: Path) -> None:
        """
        Setup data.
        """
        with self._lock:
            if (
                self._global_cache_dir is not None
                and self._global_cache_dir != global_cache_dir
            ):
                raise ValueError(
                    "Setup mismatch: global cache dir cannot be both"
                    f" {self._global_cache_dir} and {global_cache_dir}"
                )
            self._global_cache_dir = global_cache_dir

    def clear(self) -> None:
        """
        Clear collection cache.
        """
        with self._lock:
            self._global_collection_list = None
            self._global_collection_list_per_ansible_core_version.clear()

    def get_cached(
        self, *, ansible_core_version: AnsibleCoreVersion | None = None
    ) -> CollectionList | None:
        """
        Return cached list of collections, if present.
        Do not modify the result!
        """
        if ansible_core_version is None:
            return self._global_collection_list
        return self._global_collection_list_per_ansible_core_version.get(
            ansible_core_version
        )

    def get(
        self, *, ansible_core_version: AnsibleCoreVersion, runner: Runner
    ) -> CollectionList:
        """
        Search for a list of collections. The result is cached.
        """
        with self._lock:
            if self._global_cache_dir is None:
                raise ValueError("Internal error: global cache dir not setup")
            global_list = self._global_collection_list
            if global_list is None:
                global_list = CollectionList.collect_global(runner=runner)
                self._global_collection_list = global_list
            local_list = self._global_collection_list_per_ansible_core_version.get(
                ansible_core_version
            )
            if local_list is None:
                local_list = CollectionList.collect_local(
                    global_cache=_GlobalCache.create(root=self._global_cache_dir),
                    ansible_core_version=ansible_core_version,
                    current=global_list.current,
                )
                self._global_collection_list_per_ansible_core_version[
                    ansible_core_version
                ] = local_list
        return global_list.merge_with(local_list)

    def _get_global_cache(self) -> _GlobalCache:
        """
        Returns the global cache dir.
        """
        if self._global_cache_dir is None:
            raise ValueError("Internal error: global cache dir not setup")
        return _GlobalCache.create(root=self._global_cache_dir)

    def _add_collection(
        self,
        *,
        directory: Path,
        namespace: str,
        name: str,
        ansible_core_version: AnsibleCoreVersion,
    ) -> CollectionData:
        """
        Add collection in directory if the collection list has been cached.
        """
        local_list = self._global_collection_list_per_ansible_core_version.get(
            ansible_core_version
        )
        if not local_list:
            raise ValueError(
                f"Internal error: collections not listed for {ansible_core_version}"
            )
        data = load_collection_data_from_disk(directory, namespace=namespace, name=name)
        local_list._add(data)  # pylint: disable=protected-access
        return data

    @contextmanager
    def _update_collection_list(
        self, *, ansible_core_version: AnsibleCoreVersion
    ) -> t.Iterator[_CollectionListUpdater]:
        with self._lock:
            global_list = self._global_collection_list
            local_list = self._global_collection_list_per_ansible_core_version.get(
                ansible_core_version
            )
            if not global_list or self._global_cache_dir is None or local_list is None:
                raise ValueError(
                    "Internal error: collections not listed or global cache not setup"
                )
            yield _CollectionListUpdater(
                owner=self,
                merged_collection_list=global_list.merge_with(local_list),
                local_collection_list=local_list,
                ansible_core_version=ansible_core_version,
            )


_COLLECTION_LIST = _CollectionListSingleton()


@contextmanager
def _update_collection_list(
    *, ansible_core_version: AnsibleCoreVersion
) -> t.Iterator[_CollectionListUpdater]:
    # pylint: disable-next=protected-access
    with _COLLECTION_LIST._update_collection_list(
        ansible_core_version=ansible_core_version
    ) as result:
        yield result


def get_collection_list(
    *, runner: Runner, global_cache_dir: Path, ansible_core_version: AnsibleCoreVersion
) -> CollectionList:
    """
    Search for a list of collections. The result is cached.
    """
    _COLLECTION_LIST.setup(global_cache_dir=global_cache_dir)
    return _COLLECTION_LIST.get(
        runner=runner, ansible_core_version=ansible_core_version
    )


__all__ = [
    "CollectionList",
    "get_collection_list",
    "load_collection_data_from_disk",
]



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Build execution environments for testing.
"""

from __future__ import annotations

import shutil
import typing as t
from dataclasses import dataclass
from pathlib import Path

import nox

from antsibull_nox.ee_config import generate_ee_config
from antsibull_nox.paths import get_outside_temp_directory

from ..collection import CollectionData, build_collection
from ..container import get_container_engine_preference, get_preferred_container_engine
from .utils import install, register


@dataclass
class ExecutionEnvironmentData:
    """
    Information for an execution environment session.
    """

    name: str
    description: str
    config: dict[str, t.Any]
    test_playbooks: list[str]
    runtime_environment: dict[str, str] | None = None
    runtime_container_options: list[str] | None = None
    runtime_extra_vars: dict[str, str] | None = None


def build_ee_image(
    *,
    session: nox.Session,
    directory: Path,
    ee_name: str,
    collection_data: CollectionData,
    container_engine: str,
) -> str:
    """
    Build container images for execution environments.

    Args:
        session: Nox session object
        directory: Path to directory that contains execution environment definition
        ee_name: Name of execution environment
        collection_data: Collection information
        container_engine: Container runtime to use

    Returns:
        Name of successfully built container image
    """
    image_name = f"{collection_data.namespace}-{collection_data.name}-{ee_name}"
    cmd = [
        "ansible-builder",
        "build",
        "--file",
        "execution-environment.yml",
        "--tag",
        image_name,
        "--container-runtime",
        container_engine,
        "--verbosity",
        "3",
        "--context",
        str(directory),
    ]
    with session.chdir(directory):
        session.run(*cmd)  # , silent=True)
    return image_name


def prepare_execution_environment(
    *,
    session: nox.Session,
    execution_environment: ExecutionEnvironmentData,
    container_engine: str,
) -> tuple[Path | None, str | None, CollectionData]:
    """
    Generate execution environments for a collection.

    Args:
        session: Nox session object
        execution_environment: EE configuration data
        container_engine: Container runtime to use

    Returns:
        Tuple with:
        - collection_tarball_path: Path to the built collection tarball
        - built_image_names: List of built container images
        - collection_data: Collection metadata
    """
    collection_tarball_path, collection_data, _ = build_collection(session)

    if collection_tarball_path is None:
        return collection_tarball_path, None, collection_data

    directory = Path(session.create_tmp()) / "ee"
    if directory.is_dir():
        shutil.rmtree(directory)
    if not directory.is_dir():
        directory.mkdir()

    generate_ee_config(
        directory=directory,
        collection_tarball_path=collection_tarball_path.absolute(),
        collection_data=collection_data,
        ee_config=execution_environment.config,
    )

    built_image = build_ee_image(
        session=session,
        directory=directory,
        ee_name=execution_environment.name,
        collection_data=collection_data,
        container_engine=container_engine,
    )

    return collection_tarball_path, built_image, collection_data


def add_execution_environment_session(
    *,
    session_name: str,
    execution_environment: ExecutionEnvironmentData,
    default: bool = False,
    ansible_builder_package: str = "ansible-builder",
    ansible_core_package: str | None = None,
    ansible_navigator_package: str = "ansible-navigator",
) -> None:
    """
    Build and test execution environments for the collection.
    """

    def get_package_list() -> list[str]:
        result = [ansible_builder_package, ansible_navigator_package]
        if ansible_core_package is not None:
            result.append(ansible_core_package)
        return result

    def session_func(session: nox.Session) -> None:
        install(session, *get_package_list())

        container_engine = get_preferred_container_engine()
        session.log(f"Using container engine {container_engine}")

        collection_tarball, built_image, collection_data = (
            prepare_execution_environment(
                session=session,
                execution_environment=execution_environment,
                container_engine=container_engine,
            )
        )

        if collection_tarball is None or built_image is None:
            # Install only
            return

        session.log(
            f"Building execution environment {execution_environment.description}"
            f" for {collection_data.namespace}.{collection_data.name}. Image: {built_image}"
            f" using {container_engine}"
        )

        playbook_dir = Path.cwd()
        temp_dir = get_outside_temp_directory(playbook_dir.absolute())

        for playbook in execution_environment.test_playbooks:
            env = {"TMPDIR": str(temp_dir)}
            command = [
                "ansible-navigator",
                "run",
                "--mode",
                "stdout",
                "--container-engine",
                container_engine,
            ]
            if execution_environment.runtime_container_options:
                for value in execution_environment.runtime_container_options:
                    command.append(f"--container-options={value}")
            command.extend(["--pull-policy", "never"])
            if execution_environment.runtime_environment:
                for k, v in execution_environment.runtime_environment.items():
                    command.extend(["--set-environment-variable", f"{k}={v}"])
            command.extend(["--execution-environment-image", built_image])
            # Note that another parameter must follow after --set-environment-variable
            # to prevent an argument parsing SNAFU by ansible-navigator.
            # Otherwise you get errors such as "Error: The following set-environment-variable
            # entry could not be parsed: tests/ee/all.yml"...
            command.extend(
                [
                    "-v",
                    playbook,
                ]
            )
            if execution_environment.runtime_extra_vars:
                for k, v in execution_environment.runtime_extra_vars.items():
                    command.extend(["-e", f"{k}={v}"])
            session.run(
                *command,
                env=env,
            )

    # Get container engine preference to check for valid values
    get_container_engine_preference()

    session_func.__doc__ = (
        "Build and test execution environment image:"
        f" {execution_environment.description}"
    )
    nox.session(name=session_name, default=default)(session_func)

    data = {
        "name": session_name,
        "description": execution_environment.description,
    }
    register("execution-environment", data)


def add_execution_environment_sessions(
    *,
    execution_environments: list[ExecutionEnvironmentData],
    default: bool = False,
    ansible_builder_package: str = "ansible-builder",
    ansible_core_package: str | None = None,
    ansible_navigator_package: str = "ansible-navigator",
) -> None:
    """
    Build and test execution environments for the collection.
    """

    session_names = []
    for ee in execution_environments:
        session_name = f"ee-check-{ee.name}"
        add_execution_environment_session(
            session_name=session_name,
            execution_environment=ee,
            default=False,
            ansible_builder_package=ansible_builder_package,
            ansible_core_package=ansible_core_package,
            ansible_navigator_package=ansible_navigator_package,
        )
        session_names.append(session_name)

    def session_func(
        session: nox.Session,  # pylint: disable=unused-argument
    ) -> None:
        pass

    session_func.__doc__ = (
        "Meta session for building and testing execution environment images"
    )
    nox.session(
        name="ee-check",
        requires=session_names,
        default=default,
    )(session_func)



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

from __future__ import annotations

import re
import typing as t
from pathlib import Path

import pytest
from antsibull_fileutils.yaml import store_yaml_file

from antsibull_nox.ansible import (
    _CURRENT_DEVEL_VERSION,
    _CURRENT_MILESTONE_VERSION,
    _MIN_SUPPORTED_VERSION,
    AnsibleCoreVersion,
    _read_requires_ansible,
    get_ansible_core_info,
    get_ansible_core_package_name,
    get_supported_core_versions,
)
from antsibull_nox.python import _PYTHON_VERSIONS_TO_TRY
from antsibull_nox.utils import Version, version_range

from .utils import chdir


def test_check_devel_version() -> None:
    assert get_ansible_core_info("devel").ansible_core_version == _CURRENT_DEVEL_VERSION


def test_check_milestone_version() -> None:
    assert (
        get_ansible_core_info("milestone").ansible_core_version
        == _CURRENT_MILESTONE_VERSION
    )


def test_unknown_core_version() -> None:
    prev = _MIN_SUPPORTED_VERSION.previous_minor_version()
    with pytest.raises(
        ValueError, match=f"^Unknown ansible-core version {re.escape(str(prev))}$"
    ):
        get_ansible_core_info(prev)


def test_all_versions() -> None:
    # Make sure we have information on all ansible-core versions from 2.9 up to devel/milestone
    min_version = _MIN_SUPPORTED_VERSION
    max_version = max(_CURRENT_DEVEL_VERSION, _CURRENT_MILESTONE_VERSION)
    python_versions: set[Version] = set()
    for version in version_range(min_version, max_version, inclusive=True):
        info = get_ansible_core_info(version)
        assert info.ansible_core_version == version
        python_versions.update(info.controller_python_versions)
        python_versions.update(info.remote_python_versions)

    # Make sure that we know how to look for all Python versions that are in use
    for python_version in sorted(python_versions):
        assert python_version in _PYTHON_VERSIONS_TO_TRY

    # Make sure that we have all intermediate Python versions
    all_py3 = [
        python_version
        for python_version in python_versions
        if python_version.major == 3
    ]
    min_py3 = min(all_py3)
    max_py3 = max(all_py3)
    for version in version_range(min_py3, max_py3, inclusive=True):
        assert version in _PYTHON_VERSIONS_TO_TRY


GET_ANSIBLE_CORE_PAGAGE_NAME_DATA: list[
    tuple[AnsibleCoreVersion, dict[str, t.Any], str]
] = [
    (
        Version(2, 9),
        {},
        "https://github.com/ansible-community/eol-ansible/archive/stable-2.9.tar.gz",
    ),
    (
        Version(2, 9),
        {"source": "pypi"},
        "ansible>=2.9,<2.10",
    ),
    (
        Version(2, 10),
        {},
        "https://github.com/ansible-community/eol-ansible/archive/stable-2.10.tar.gz",
    ),
    (
        Version(2, 10),
        {"source": "pypi"},
        "ansible-base>=2.10,<2.11",
    ),
    (
        Version(2, 11),
        {},
        "https://github.com/ansible-community/eol-ansible/archive/stable-2.11.tar.gz",
    ),
    (
        Version(2, 11),
        {"source": "pypi"},
        "ansible-core>=2.11,<2.12",
    ),
    # Last EOL version
    (
        Version(2, 14),
        {},
        "https://github.com/ansible-community/eol-ansible/archive/stable-2.14.tar.gz",
    ),
    # First non-EOL version
    (
        Version(2, 15),
        {},
        "https://github.com/ansible/ansible/archive/stable-2.15.tar.gz",
    ),
    # devel
    (
        "devel",
        {},
        "https://github.com/ansible/ansible/archive/devel.tar.gz",
    ),
    (
        "devel",
        {"source": "pypi"},
        "https://github.com/ansible/ansible/archive/devel.tar.gz",
    ),
    # milestone
    (
        "milestone",
        {},
        "https://github.com/ansible/ansible/archive/milestone.tar.gz",
    ),
    (
        "milestone",
        {"source": "pypi"},
        "https://github.com/ansible/ansible/archive/milestone.tar.gz",
    ),
    (
        "milestone",
        {
            "branch_name": "refs/pull/84621/head",
            "ansible_repo": "ansible-community/eol-ansible",
        },
        "https://github.com/ansible-community/eol-ansible/archive/refs/pull/84621/head.tar.gz",
    ),
]


@pytest.mark.parametrize(
    "version, kwargs, expected_package_name",
    GET_ANSIBLE_CORE_PAGAGE_NAME_DATA,
)
def test_get_ansible_core_package_name(
    version: AnsibleCoreVersion, kwargs: dict[str, t.Any], expected_package_name: str
) -> None:
    assert get_ansible_core_package_name(version, **kwargs) == expected_package_name


GET_SUPPORTED_CORE_VERSIONS_DATA: list[
    tuple[str, dict[str, t.Any], list[AnsibleCoreVersion]]
] = [
    (
        ">= 2.9.10",
        {},
        [
            Version.parse("2.9"),
            Version.parse("2.10"),
            Version.parse("2.11"),
            Version.parse("2.12"),
            Version.parse("2.13"),
            Version.parse("2.14"),
            Version.parse("2.15"),
            Version.parse("2.16"),
            Version.parse("2.17"),
            Version.parse("2.18"),
            Version.parse("2.19"),
        ],
    ),
    (
        ">= 2.9.10, < 2.13",
        {"include_devel": True},
        [
            Version.parse("2.9"),
            Version.parse("2.10"),
            Version.parse("2.11"),
            Version.parse("2.12"),
            "devel",
        ],
    ),
    (
        ">= 2.9.10, < 2.16",
        {
            "min_version": Version.parse("2.11"),
            "max_version": Version.parse("2.14"),
            "except_versions": (Version.parse("2.13"),),
            "include_milestone": True,
            "include_devel": True,
        },
        [
            Version.parse("2.11"),
            Version.parse("2.12"),
            Version.parse("2.14"),
            "milestone",
            "devel",
        ],
    ),
    (
        ">= 2.9.10, < 2.11",
        {
            "except_versions": (Version.parse("2.9"), "milestone", "devel"),
            "include_milestone": True,
            "include_devel": True,
        },
        [
            Version.parse("2.10"),
        ],
    ),
]


def test__read_requires_ansible_fail(tmp_path: Path) -> None:
    meta = tmp_path / "meta"
    meta.mkdir()
    runtime_yml = meta / "runtime.yml"
    with chdir(tmp_path):
        _read_requires_ansible.cache_clear()  # added by @functools.cache
        with pytest.raises(ValueError, match="^Cannot open meta/runtime.yml$"):
            _read_requires_ansible()

        _read_requires_ansible.cache_clear()  # added by @functools.cache
        runtime_yml.mkdir()
        with pytest.raises(
            ValueError,
            match=(
                r"^Cannot parse meta/runtime.yml: \[Errno 21\]"
                " Is a directory: 'meta/runtime.yml'$"
            ),
        ):
            _read_requires_ansible()
        runtime_yml.rmdir()

        runtime_yml.write_text("{")
        _read_requires_ansible.cache_clear()  # added by @functools.cache
        with pytest.raises(
            ValueError,
            match="^Cannot parse meta/runtime.yml: while parsing a flow node",
        ):
            _read_requires_ansible()

        store_yaml_file(runtime_yml, {"requires_ansible": 123})
        _read_requires_ansible.cache_clear()  # added by @functools.cache
        with pytest.raises(
            ValueError,
            match="^meta/runtime.yml does not contain an 'requires_ansible' string$",
        ):
            _read_requires_ansible()

        store_yaml_file(runtime_yml, {"requires_ansible": []})
        _read_requires_ansible.cache_clear()  # added by @functools.cache
        with pytest.raises(
            ValueError,
            match="^meta/runtime.yml does not contain an 'requires_ansible' string$",
        ):
            _read_requires_ansible()

        store_yaml_file(runtime_yml, {})
        _read_requires_ansible.cache_clear()  # added by @functools.cache
        with pytest.raises(
            ValueError,
            match="^meta/runtime.yml does not contain an 'requires_ansible' string$",
        ):
            _read_requires_ansible()

        store_yaml_file(runtime_yml, {"requires_ansible": "<>"})
        _read_requires_ansible.cache_clear()  # added by @functools.cache
        with pytest.raises(
            ValueError,
            match=(
                "^meta/runtime.yml contains an invalid"
                " 'requires_ansible' string: Invalid specifier: '<>'$"
            ),
        ):
            _read_requires_ansible()


@pytest.mark.parametrize(
    "requires_ansible, kwargs, expected_version_list",
    GET_SUPPORTED_CORE_VERSIONS_DATA,
)
def test_get_supported_core_versions(
    requires_ansible: str,
    kwargs: dict[str, t.Any],
    expected_version_list: list[AnsibleCoreVersion],
    tmp_path: Path,
) -> None:
    meta = tmp_path / "meta"
    meta.mkdir()
    runtime_yml = meta / "runtime.yml"
    store_yaml_file(runtime_yml, {"requires_ansible": requires_ansible})
    with chdir(tmp_path):
        _read_requires_ansible.cache_clear()  # added by @functools.cache
        get_supported_core_versions.cache_clear()  # added by @functools.cache
        print(kwargs)
        version_list = get_supported_core_versions(**kwargs)
    assert version_list == expected_version_list



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Create nox docs check session.
"""

from __future__ import annotations

import typing as t
from pathlib import Path

import nox

from ..paths import (
    list_all_files,
)
from .collections import (
    CollectionSetup,
    prepare_collections,
)
from .utils import (
    get_package_version,
    install,
    is_new_enough,
    run_bare_script,
)


def find_extra_docs_rst_files() -> list[Path]:
    """
    Find all RST extra document files.
    """
    all_files = list_all_files()
    cwd = Path.cwd()
    extra_docs_dir = cwd / "docs" / "docsite" / "rst"
    return [
        file
        for file in all_files
        if file.is_relative_to(extra_docs_dir) and file.name.lower().endswith((".rst"))
    ]


def add_docs_check(
    *,
    make_docs_check_default: bool = True,
    antsibull_docs_package: str = "antsibull-docs",
    ansible_core_package: str = "ansible-core",
    validate_collection_refs: t.Literal["self", "dependent", "all"] | None = None,
    extra_collections: list[str] | None = None,
    codeblocks_restrict_types: list[str] | None = None,
    codeblocks_restrict_type_exact_case: bool = True,
    codeblocks_allow_without_type: bool = True,
    codeblocks_allow_literal_blocks: bool = True,
    antsibull_docutils_package: str = "antsibull-docutils",
) -> None:
    """
    Add docs-check session for linting.
    """
    run_extra_checks = (
        codeblocks_restrict_types is not None
        or not codeblocks_allow_without_type
        or not codeblocks_allow_literal_blocks
    )

    def compose_dependencies() -> list[str]:
        deps = [antsibull_docs_package, ansible_core_package]
        if run_extra_checks:
            deps.append(antsibull_docutils_package)
        return deps

    def execute_extra_checks(session: nox.Session) -> None:
        all_extra_docs = find_extra_docs_rst_files()
        if not all_extra_docs:
            session.warn(
                "Skipping codeblock checks since no appropriate RST file was found..."
            )
            return
        run_bare_script(
            session,
            "rst-extra",
            use_session_python=True,
            files=all_extra_docs,
            extra_data={
                "codeblocks_restrict_types": codeblocks_restrict_types,
                "codeblocks_restrict_type_exact_case": codeblocks_restrict_type_exact_case,
                "codeblocks_allow_without_type": codeblocks_allow_without_type,
                "codeblocks_allow_literal_blocks": codeblocks_allow_literal_blocks,
            },
        )

    def execute_antsibull_docs(
        session: nox.Session, prepared_collections: CollectionSetup
    ) -> None:
        antsibull_docs_version = get_package_version(session, "antsibull-docs")
        if antsibull_docs_version is not None:
            session.log(f"Detected antsibull-docs version {antsibull_docs_version}")
        with session.chdir(prepared_collections.current_path):
            collections_path = f"{prepared_collections.current_place}"
            command = [
                "antsibull-docs",
                "lint-collection-docs",
                "--plugin-docs",
                "--skip-rstcheck",
                ".",
            ]
            if validate_collection_refs:
                command.extend(["--validate-collection-refs", validate_collection_refs])
            if is_new_enough(antsibull_docs_version, min_version="2.18.0"):
                command.append("--check-extra-docs-refs")
            session.run(*command, env={"ANSIBLE_COLLECTIONS_PATH": collections_path})

    def docs_check(session: nox.Session) -> None:
        install(session, *compose_dependencies())
        prepared_collections = prepare_collections(
            session,
            install_in_site_packages=False,
            extra_collections=extra_collections,
            install_out_of_tree=True,
        )
        if run_extra_checks:
            execute_extra_checks(session)
        if not prepared_collections:
            session.warn("Skipping antsibull-docs...")
        if prepared_collections:
            execute_antsibull_docs(session, prepared_collections)

    docs_check.__doc__ = "Run 'antsibull-docs lint-collection-docs'"
    nox.session(
        name="docs-check",
        default=make_docs_check_default,
    )(docs_check)


__all__ = [
    "add_docs_check",
]



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt
# or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025 Felix Fontein <felix@fontein.de>

# /// script
# dependencies = ["nox>=2025.02.09", "antsibull-nox"]
# ///

import sys

import nox

try:
    import antsibull_nox
except ImportError:
    print("You need to install antsibull-nox in the same Python environment as nox.")
    sys.exit(1)


antsibull_nox.load_antsibull_nox_toml()


# Allow to run the noxfile with `python noxfile.py`, `pipx run noxfile.py`, or similar.
# Requires nox >= 2025.02.09
if __name__ == "__main__":
    nox.main()



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Utils for creating nox sessions.
"""

from __future__ import annotations

import json
import logging
import os
import sys
import typing as t
from contextlib import contextmanager
from pathlib import Path

import nox
from nox.logger import OUTPUT as nox_OUTPUT
from packaging.version import InvalidVersion
from packaging.version import parse as parse_version

from ..data_util import prepare_data_script
from ..paths import (
    find_data_directory,
    list_all_files,
)

# https://docs.github.com/en/actions/writing-workflows/choosing-what-your-workflow-does/store-information-in-variables#default-environment-variables
# https://docs.gitlab.com/ci/variables/predefined_variables/#predefined-variables
# https://docs.travis-ci.com/user/environment-variables/#default-environment-variables
IN_CI = os.environ.get("CI") == "true"
IN_GITHUB_ACTIONS = bool(os.environ.get("GITHUB_ACTION"))
ALLOW_EDITABLE = os.environ.get("ALLOW_EDITABLE", str(not IN_CI)).lower() in (
    "1",
    "true",
)

_SESSIONS: dict[str, list[dict[str, t.Any]]] = {}


def nox_has_verbosity() -> bool:
    """
    Determine whether nox is run with verbosity enabled.
    """
    logger = logging.getLogger()
    return logger.level <= nox_OUTPUT


@contextmanager
def silence_run_verbosity() -> t.Iterator[None]:
    """
    When using session.run() with silent=True, nox will log the output
    if -v is used. Using this context manager prevents printing the output.
    """
    logger = logging.getLogger()
    original_level = logger.level
    try:
        logger.setLevel(max(nox_OUTPUT + 1, original_level))
        yield
    finally:
        logger.setLevel(original_level)


@contextmanager
def ci_group(name: str) -> t.Iterator[tuple[str, bool]]:
    """
    Try to ensure that the output inside the context is printed in a collapsable group.

    This is highly CI system dependent, and currently only works for GitHub Actions.
    """
    is_collapsing = False
    if IN_GITHUB_ACTIONS:
        print(f"::group::{name}")
        sys.stdout.flush()
        is_collapsing = True
    yield ("  " if is_collapsing else "", is_collapsing)
    if IN_GITHUB_ACTIONS:
        print("::endgroup::")
        sys.stdout.flush()


def register(name: str, data: dict[str, t.Any]) -> None:
    """
    Register a session name for matrix generation with additional data.
    """
    if name not in _SESSIONS:
        _SESSIONS[name] = []
    _SESSIONS[name].append(data)


def get_registered_sessions() -> dict[str, list[dict[str, t.Any]]]:
    """
    Return all registered sessions.
    """
    return {
        name: [session.copy() for session in sessions]
        for name, sessions in _SESSIONS.items()
    }


def install(session: nox.Session, *args: str, editable: bool = False, **kwargs):
    """
    Install Python packages.
    """
    if not args:
        return
    # nox --no-venv
    if isinstance(session.virtualenv, nox.virtualenv.PassthroughEnv):
        session.warn(f"No venv. Skipping installation of {args}")
        return
    # Don't install in editable mode in CI or if it's explicitly disabled.
    # This ensures that the wheel contains all of the correct files.
    if editable and ALLOW_EDITABLE:
        args = ("-e", *args)
    session.install(*args, "-U", **kwargs)


def run_bare_script(
    session: nox.Session,
    /,
    name: str,
    *,
    use_session_python: bool = False,
    files: list[Path] | None = None,
    extra_data: dict[str, t.Any] | None = None,
) -> None:
    """
    Run a bare script included in antsibull-nox's data directory.
    """
    if files is None:
        files = list_all_files()
    data = prepare_data_script(
        session,
        base_name=name,
        paths=files,
        extra_data=extra_data,
    )
    python = sys.executable
    env = {}
    if use_session_python:
        python = "python"
        env["PYTHONPATH"] = str(find_data_directory())
    session.run(
        python,
        find_data_directory() / f"{name}.py",
        "--data",
        data,
        external=True,
        env=env,
    )


def get_package_versions(
    session: nox.Session,
    /,
    packages: list[str] | str,
    *,
    use_session_python: bool = True,
) -> None | dict[str, str | None]:
    """
    Retrieve the versions of one or more Python packages.
    """
    name = "get-package-versions"
    if isinstance(packages, str):
        packages = [packages]
    if not packages:
        return {}
    data = prepare_data_script(
        session,
        base_name=name,
        paths=[],
        extra_data={
            "packages": packages,
        },
    )
    python = sys.executable
    env = {}
    if use_session_python:
        python = "python"
        env["PYTHONPATH"] = str(find_data_directory())
    result = session.run(
        python,
        find_data_directory() / f"{name}.py",
        "--data",
        data,
        external=True,
        silent=True,
        env=env,
    )
    if result is None:
        return None
    return json.loads(result)


def get_package_version(
    session: nox.Session,
    /,
    package: str,
    *,
    use_session_python: bool = True,
) -> str | None:
    """
    Retrieve a Python package's version.
    """
    result = get_package_versions(
        session, package, use_session_python=use_session_python
    )
    return None if result is None else result.get(package)


def is_new_enough(actual_version: str | None, *, min_version: str) -> bool:
    """
    Given a program version, compares it to the min_version.
    If the program version is not given, it is assumed to be "new enough".
    """
    if actual_version is None:
        return True
    try:
        act_v = parse_version(actual_version)
    except InvalidVersion as exc:
        raise ValueError(
            f"Cannot parse actual version {actual_version!r}: {exc}"
        ) from exc
    try:
        min_v = parse_version(min_version)
    except InvalidVersion as exc:
        raise ValueError(
            f"Cannot parse minimum version {min_version!r}: {exc}"
        ) from exc
    return act_v >= min_v


def compose_description(
    *,
    prefix: str | dict[t.Literal["one", "other"], str] | None = None,
    programs: dict[str, str | bool | None],
) -> str:
    """
    Compose a description for a nox session from several configurable parts.
    """
    parts: list[str] = []

    def add(text: str, *, comma: bool = False) -> None:
        if parts:
            if comma:
                parts.append(", ")
            else:
                parts.append(" ")
        parts.append(text)

    active_programs = [
        (program, value if isinstance(value, str) else None)
        for program, value in programs.items()
        if value not in (False, None)
    ]

    if prefix:
        if isinstance(prefix, dict):
            if len(active_programs) == 1 and "one" in prefix:
                add(prefix["one"])
            else:
                add(prefix["other"])
        else:
            add(prefix)

    for index, (program, value) in enumerate(active_programs):
        if index + 1 == len(active_programs) and index > 0:
            add("and", comma=index > 1)
        add(program, comma=index > 0 and index + 1 < len(active_programs))
        if value is not None:
            add(f"({value})")

    return "".join(parts)


__all__ = [
    "ci_group",
    "compose_description",
    "get_package_version",
    "get_package_versions",
    "get_registered_sessions",
    "install",
    "is_new_enough",
    "nox_has_verbosity",
    "register",
    "run_bare_script",
    "silence_run_verbosity",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Utility code for scripts in data.
"""

from __future__ import annotations

import json
import typing as t
from pathlib import Path

import nox


def prepare_data_script(
    session: nox.Session,
    *,
    base_name: str,
    paths: list[Path],
    extra_data: dict[str, t.Any] | None = None,
) -> Path:
    """
    Prepare a data JSON file for the extra sanity check scripts.
    """
    cwd = Path.cwd()
    data = {}
    data["paths"] = [str(path.relative_to(cwd)) for path in paths]
    if extra_data:
        data.update(extra_data)
    file = Path(session.create_tmp()) / f"{base_name}-data.json"
    with open(file, "w", encoding="utf-8") as f:
        json.dump(data, f)
    return file



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Create nox license check session.
"""

from __future__ import annotations

import nox

from .utils import (
    compose_description,
    install,
    run_bare_script,
)


def add_license_check(
    *,
    make_license_check_default: bool = True,
    run_reuse: bool = True,
    reuse_package: str = "reuse",
    run_license_check: bool = True,
    license_check_extra_ignore_paths: list[str] | None = None,
) -> None:
    """
    Add license-check session for license checks.
    """

    def compose_dependencies() -> list[str]:
        deps = []
        if run_reuse:
            deps.append(reuse_package)
        return deps

    def license_check(session: nox.Session) -> None:
        install(session, *compose_dependencies())
        if run_reuse:
            session.run("reuse", "lint")
        if run_license_check:
            run_bare_script(
                session,
                "license-check",
                extra_data={
                    "extra_ignore_paths": license_check_extra_ignore_paths or [],
                },
            )

    license_check.__doc__ = compose_description(
        prefix={
            "one": "Run license checker:",
            "other": "Run license checkers:",
        },
        programs={
            "reuse": run_reuse,
            "license-check": (
                "ensure GPLv3+ for plugins" if run_license_check else False
            ),
        },
    )
    nox.session(
        name="license-check",
        default=make_license_check_default,
    )(license_check)


__all__ = [
    "add_license_check",
]



#!/usr/bin/env python

# Copyright (c) 2025, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt
# or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""Make sure all plugin and module documentation adheres to yamllint."""

from __future__ import annotations

import os
import sys
import traceback
import typing as t

from antsibull_docutils.rst_code_finder import find_code_blocks
from antsibull_nox_data_util import get_bool, get_list_of_strings, setup  # type: ignore


def process_rst_file(
    errors: list[dict[str, t.Any]],
    path: str,
    *,
    codeblocks_restrict_types: list[str] | None,
    codeblocks_restrict_type_exact_case: bool,
    codeblocks_allow_without_type: bool,
    codeblocks_allow_literal_blocks: bool,
) -> None:
    try:
        with open(path, "rt", encoding="utf-8") as f:
            content = f.read()
    except Exception as exc:
        errors.append(
            {
                "path": path,
                "line": 1,
                "col": 1,
                "message": (
                    f"Error while reading content: {type(exc)}:"
                    f" {exc}; traceback: {traceback.format_exc()!r}"
                ),
            }
        )
        return

    def warn_unknown_block(line: int | str, col: int, content: str) -> None:
        if not codeblocks_allow_literal_blocks:
            errors.append(
                {
                    "path": path,
                    "line": line,
                    "col": col,
                    "message": (
                        "Warning: found unknown literal block! Check for double colons '::'."
                        " If that is not the cause, please report this warning."
                        " It might indicate a bug in the checker"
                        " or an unsupported Sphinx directive."
                        f" Content: {content!r}"
                    ),
                }
            )

    for code_block in find_code_blocks(
        content,
        path=path,
        root_prefix="docs/docsite/rst",
        warn_unknown_block=warn_unknown_block,
    ):
        error_data = {
            "path": path,
            "line": code_block.row_offset + 1,
            "col": code_block.col_offset + 1,
        }
        if not code_block.position_exact:
            error_data["note"] = (
                "The code block could not be exactly located in the source file."
                " The line/column numbers might be off."
            )

        if code_block.language is None:
            if not codeblocks_allow_without_type:
                msg = "Every code block must have a language."
                if codeblocks_restrict_types is not None:
                    langs = ", ".join(sorted(codeblocks_restrict_types))
                    msg = f"{msg} Allowed languages are: {langs}"
                error_data["message"] = msg
                errors.append(error_data)
            continue

        if codeblocks_restrict_types is None:
            continue

        language = code_block.language
        if not codeblocks_restrict_type_exact_case:
            language = language.lower()

        if language not in codeblocks_restrict_types:
            langs = ", ".join(sorted(codeblocks_restrict_types))
            msg = (
                f"Code block with disallowed language {code_block.language!r} found."
                f" Allowed languages are: {langs}"
            )
            error_data["message"] = msg
            errors.append(error_data)


def main() -> int:
    """Main entry point."""
    paths, extra_data = setup()
    codeblocks_restrict_types = get_list_of_strings(
        extra_data, "codeblocks_restrict_types"
    )
    codeblocks_restrict_type_exact_case = get_bool(
        extra_data, "codeblocks_restrict_type_exact_case", default=True
    )
    codeblocks_allow_without_type = get_bool(
        extra_data, "codeblocks_allow_without_type", default=True
    )
    codeblocks_allow_literal_blocks = get_bool(
        extra_data, "codeblocks_allow_literal_blocks", default=True
    )

    if (
        codeblocks_restrict_types is not None
        and not codeblocks_restrict_type_exact_case
    ):
        codeblocks_restrict_types = [
            language.lower() for language in codeblocks_restrict_types
        ]

    errors: list[dict[str, t.Any]] = []
    for path in paths:
        if not os.path.isfile(path):
            continue
        process_rst_file(
            errors,
            path,
            codeblocks_restrict_types=codeblocks_restrict_types,
            codeblocks_restrict_type_exact_case=codeblocks_restrict_type_exact_case,
            codeblocks_allow_without_type=codeblocks_allow_without_type,
            codeblocks_allow_literal_blocks=codeblocks_allow_literal_blocks,
        )

    errors.sort(
        key=lambda error: (
            error["path"],
            error["line"] if isinstance(error["line"], int) else 0,
            error["col"],
            error["message"],
        )
    )
    for error in errors:
        prefix = f"{error['path']}:{error['line']}:{error['col']}: "
        msg = error["message"]
        if "note" in error:
            msg = f"{msg}\nNote: {error['note']}"
        for i, line in enumerate(msg.splitlines()):
            print(f"{prefix}{line}")
            if i == 0:
                prefix = " " * len(prefix)

    return len(errors) > 0


if __name__ == "__main__":
    sys.exit(main())



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Create nox ansible-test sessions.
"""

from __future__ import annotations

import os
import typing as t
from collections.abc import Callable
from pathlib import Path

import nox
from antsibull_fileutils.yaml import store_yaml_file

from ..ansible import (
    AnsibleCoreVersion,
    get_ansible_core_info,
    get_ansible_core_package_name,
    get_supported_core_versions,
    parse_ansible_core_version,
)
from ..container import get_container_engine_preference
from ..paths import copy_directory_tree_into
from ..python import get_installed_python_versions
from ..utils import Version
from .collections import prepare_collections
from .utils import (
    install,
    register,
)


def get_ansible_test_env() -> dict[str, str]:
    """
    Set ANSIBLE_TEST_PREFER_PODMAN if the user prefers one of podman or docker over the other.
    """
    # Check whether the user explicitly set ANSIBLE_TEST_PREFER_PODMAN
    if os.environ.get("ANSIBLE_TEST_PREFER_PODMAN") is not None:
        return {}
    # Check whether the user explicitly requested a container engine for antsibull-nox
    preference, explicitly_set = get_container_engine_preference()
    if not explicitly_set:
        return {}
    # Check whether the users prefers one of podman and docker over the other
    if preference in ("auto-prefer-podman", "podman"):
        # Yes, the user prefers podman.
        return {"ANSIBLE_TEST_PREFER_PODMAN": "1"}
    if preference in ("auto-prefer-docker", "docker"):
        # Yes, the user prefers docker.
        return {"ANSIBLE_TEST_PREFER_PODMAN": ""}
    # Apparently not: do nothing.
    return {}


def add_ansible_test_session(
    *,
    name: str,
    description: str | None,
    extra_deps_files: list[str | os.PathLike] | None = None,
    ansible_test_params: list[str],
    add_posargs: bool = True,
    default: bool,
    ansible_core_version: str | AnsibleCoreVersion,
    ansible_core_source: t.Literal["git", "pypi"] = "git",
    ansible_core_repo_name: str | None = None,
    ansible_core_branch_name: str | None = None,
    handle_coverage: t.Literal["never", "always", "auto"] = "auto",
    register_name: str | None = None,
    register_extra_data: dict[str, t.Any] | None = None,
    callback_before: Callable[[], None] | None = None,
    callback_after: Callable[[], None] | None = None,
) -> None:
    """
    Add generic ansible-test session.

    Returns a list of Python versions set for this session.
    """
    parsed_ansible_core_version = parse_ansible_core_version(ansible_core_version)

    def compose_dependencies() -> list[str]:
        deps = [
            get_ansible_core_package_name(
                parsed_ansible_core_version,
                source=ansible_core_source,
                ansible_repo=ansible_core_repo_name,
                branch_name=ansible_core_branch_name,
            )
        ]
        return deps

    def run_ansible_test(session: nox.Session) -> None:
        install(session, *compose_dependencies())
        prepared_collections = prepare_collections(
            session,
            ansible_core_version=parsed_ansible_core_version,
            install_in_site_packages=False,
            extra_deps_files=extra_deps_files,
            install_out_of_tree=True,
        )
        if not prepared_collections:
            session.warn("Skipping ansible-test...")
            return
        cwd = Path.cwd()
        with session.chdir(prepared_collections.current_path):
            if callback_before:
                callback_before()

            command = ["ansible-test"] + ansible_test_params
            if add_posargs and session.posargs:
                command.extend(session.posargs)
            session.run(*command, env=get_ansible_test_env())

            coverage = (handle_coverage == "auto" and "--coverage" in command) or (
                handle_coverage == "always"
            )
            if coverage:
                session.run(
                    "ansible-test",
                    "coverage",
                    "xml",
                    "--color",
                    "-v",
                    "--requirements",
                    "--group-by",
                    "command",
                    "--group-by",
                    "version",
                )

            if callback_after:
                callback_after()

            copy_directory_tree_into(
                prepared_collections.current_path / "tests" / "output",
                cwd / "tests" / "output",
            )

    # Determine Python version(s)
    core_info = get_ansible_core_info(parsed_ansible_core_version)
    all_versions = get_installed_python_versions()

    installed_versions = [
        version
        for version in core_info.controller_python_versions
        if version in all_versions
    ]
    python = max(installed_versions or core_info.controller_python_versions)
    python_versions = [python]

    run_ansible_test.__doc__ = description
    nox.session(
        name=name,
        default=default,
        python=[str(python_version) for python_version in python_versions],
    )(run_ansible_test)

    if register_name:
        data = {
            "name": name,
            "ansible-core": (
                str(ansible_core_branch_name)
                if ansible_core_branch_name is not None
                else str(parsed_ansible_core_version)
            ),
            "python": " ".join(str(python) for python in python_versions),
        }
        if register_extra_data:
            data.update(register_extra_data)
        register(register_name, data)


def add_ansible_test_sanity_test_session(
    *,
    name: str,
    description: str | None,
    default: bool,
    ansible_core_version: str | AnsibleCoreVersion,
    ansible_core_source: t.Literal["git", "pypi"] = "git",
    ansible_core_repo_name: str | None = None,
    ansible_core_branch_name: str | None = None,
    skip_tests: list[str] | None = None,
    allow_disabled: bool = False,
    enable_optional_errors: bool = False,
    register_extra_data: dict[str, t.Any] | None = None,
) -> None:
    """
    Add generic ansible-test sanity test session.
    """
    command = ["sanity", "--color", "-v", "--docker"]
    if skip_tests:
        for test in skip_tests:
            command.extend(["--skip", test])
    if allow_disabled:
        command.append("--allow-disabled")
    if enable_optional_errors:
        command.append("--enable-optional-errors")
    add_ansible_test_session(
        name=name,
        description=description,
        ansible_test_params=command,
        default=default,
        ansible_core_version=ansible_core_version,
        ansible_core_source=ansible_core_source,
        ansible_core_repo_name=ansible_core_repo_name,
        ansible_core_branch_name=ansible_core_branch_name,
        register_name="sanity",
        register_extra_data=register_extra_data,
    )


def _parse_min_max_except(
    min_version: Version | str | None,
    max_version: Version | str | None,
    except_versions: list[AnsibleCoreVersion | str] | None,
) -> tuple[Version | None, Version | None, tuple[AnsibleCoreVersion, ...] | None]:
    if isinstance(min_version, str):
        min_version = Version.parse(min_version)
    if isinstance(max_version, str):
        max_version = Version.parse(max_version)
    if except_versions is None:
        return min_version, max_version, None
    evs = tuple(parse_ansible_core_version(version) for version in except_versions)
    return min_version, max_version, evs


def add_all_ansible_test_sanity_test_sessions(
    *,
    default: bool = False,
    include_devel: bool = False,
    include_milestone: bool = False,
    add_devel_like_branches: list[tuple[str | None, str]] | None = None,
    min_version: Version | str | None = None,
    max_version: Version | str | None = None,
    except_versions: list[AnsibleCoreVersion | str] | None = None,
    skip_tests: list[str] | None = None,
    allow_disabled: bool = False,
    enable_optional_errors: bool = False,
) -> None:
    """
    Add ansible-test sanity test meta session that runs ansible-test sanity
    for all supported ansible-core versions.
    """
    parsed_min_version, parsed_max_version, parsed_except_versions = (
        _parse_min_max_except(min_version, max_version, except_versions)
    )

    sanity_sessions = []
    for ansible_core_version in get_supported_core_versions(
        include_devel=include_devel,
        include_milestone=include_milestone,
        min_version=parsed_min_version,
        max_version=parsed_max_version,
        except_versions=parsed_except_versions,
    ):
        name = f"ansible-test-sanity-{ansible_core_version}"
        add_ansible_test_sanity_test_session(
            name=name,
            description=f"Run sanity tests from ansible-core {ansible_core_version}'s ansible-test",
            ansible_core_version=ansible_core_version,
            skip_tests=skip_tests,
            allow_disabled=allow_disabled,
            enable_optional_errors=enable_optional_errors,
            register_extra_data={"display-name": f"Ⓐ{ansible_core_version}"},
            default=False,
        )
        sanity_sessions.append(name)
    if add_devel_like_branches:
        for repo_name, branch_name in add_devel_like_branches:
            repo_prefix = (
                f"{repo_name.replace('/', '-')}-" if repo_name is not None else ""
            )
            repo_postfix = f", {repo_name} repository" if repo_name is not None else ""
            name = f"ansible-test-sanity-{repo_prefix}{branch_name.replace('/', '-')}"
            add_ansible_test_sanity_test_session(
                name=name,
                description=(
                    "Run sanity tests from ansible-test in ansible-core's"
                    f" {branch_name} branch{repo_postfix}"
                ),
                ansible_core_version="devel",
                ansible_core_repo_name=repo_name,
                ansible_core_branch_name=branch_name,
                skip_tests=skip_tests,
                allow_disabled=allow_disabled,
                enable_optional_errors=enable_optional_errors,
                register_extra_data={"display-name": f"Ⓐ{branch_name}"},
                default=False,
            )
            sanity_sessions.append(name)

    def run_all_sanity_tests(
        session: nox.Session,  # pylint: disable=unused-argument
    ) -> None:
        pass

    run_all_sanity_tests.__doc__ = (
        "Meta session for running all ansible-test-sanity-* sessions."
    )
    nox.session(
        name="ansible-test-sanity",
        default=default,
        requires=sanity_sessions,
    )(run_all_sanity_tests)


def add_ansible_test_unit_test_session(
    *,
    name: str,
    description: str | None,
    default: bool,
    ansible_core_version: str | AnsibleCoreVersion,
    ansible_core_source: t.Literal["git", "pypi"] = "git",
    ansible_core_repo_name: str | None = None,
    ansible_core_branch_name: str | None = None,
    register_extra_data: dict[str, t.Any] | None = None,
) -> None:
    """
    Add generic ansible-test unit test session.
    """
    add_ansible_test_session(
        name=name,
        description=description,
        ansible_test_params=["units", "--color", "-v", "--docker"],
        extra_deps_files=["tests/unit/requirements.yml"],
        default=default,
        ansible_core_version=ansible_core_version,
        ansible_core_source=ansible_core_source,
        ansible_core_repo_name=ansible_core_repo_name,
        ansible_core_branch_name=ansible_core_branch_name,
        register_name="units",
        register_extra_data=register_extra_data,
    )


def add_all_ansible_test_unit_test_sessions(
    *,
    default: bool = False,
    include_devel: bool = False,
    include_milestone: bool = False,
    add_devel_like_branches: list[tuple[str | None, str]] | None = None,
    min_version: Version | str | None = None,
    max_version: Version | str | None = None,
    except_versions: list[AnsibleCoreVersion | str] | None = None,
) -> None:
    """
    Add ansible-test unit test meta session that runs ansible-test units
    for all supported ansible-core versions.
    """
    parsed_min_version, parsed_max_version, parsed_except_versions = (
        _parse_min_max_except(min_version, max_version, except_versions)
    )

    units_sessions = []
    for ansible_core_version in get_supported_core_versions(
        include_devel=include_devel,
        include_milestone=include_milestone,
        min_version=parsed_min_version,
        max_version=parsed_max_version,
        except_versions=parsed_except_versions,
    ):
        name = f"ansible-test-units-{ansible_core_version}"
        add_ansible_test_unit_test_session(
            name=name,
            description=f"Run unit tests with ansible-core {ansible_core_version}'s ansible-test",
            ansible_core_version=ansible_core_version,
            register_extra_data={"display-name": f"Ⓐ{ansible_core_version}"},
            default=False,
        )
        units_sessions.append(name)
    if add_devel_like_branches:
        for repo_name, branch_name in add_devel_like_branches:
            repo_prefix = (
                f"{repo_name.replace('/', '-')}-" if repo_name is not None else ""
            )
            repo_postfix = f", {repo_name} repository" if repo_name is not None else ""
            name = f"ansible-test-units-{repo_prefix}{branch_name.replace('/', '-')}"
            add_ansible_test_unit_test_session(
                name=name,
                description=(
                    "Run unit tests from ansible-test in ansible-core's"
                    f" {branch_name} branch{repo_postfix}"
                ),
                ansible_core_version="devel",
                ansible_core_repo_name=repo_name,
                ansible_core_branch_name=branch_name,
                register_extra_data={"display-name": f"Ⓐ{branch_name}"},
                default=False,
            )
            units_sessions.append(name)

    def run_all_unit_tests(
        session: nox.Session,  # pylint: disable=unused-argument
    ) -> None:
        pass

    run_all_unit_tests.__doc__ = (
        "Meta session for running all ansible-test-units-* sessions."
    )
    nox.session(
        name="ansible-test-units",
        default=default,
        requires=units_sessions,
    )(run_all_unit_tests)


def add_ansible_test_integration_sessions_default_container(
    *,
    include_devel: bool = False,
    include_milestone: bool = False,
    add_devel_like_branches: list[tuple[str | None, str]] | None = None,
    min_version: Version | str | None = None,
    max_version: Version | str | None = None,
    except_versions: list[AnsibleCoreVersion | str] | None = None,
    core_python_versions: (
        dict[str | AnsibleCoreVersion, list[str | Version]] | None
    ) = None,
    controller_python_versions_only: bool = False,
    ansible_vars_from_env_vars: dict[str, str] | None = None,
    default: bool = False,
) -> None:
    """
    Add ansible-test integration tests using the default Docker container.

    ``core_python_versions`` can be used to restrict the Python versions
    to be used for a specific ansible-core version.

    ``controller_python_versions_only`` can be used to only run against
    controller Python versions.
    """

    def callback_before() -> None:
        if not ansible_vars_from_env_vars:
            return

        path = Path("tests", "integration", "integration_config.yml")
        content: dict[str, t.Any] = {}
        for ans_var, env_var in ansible_vars_from_env_vars.items():
            value = os.environ.get(env_var)
            if value is not None:
                content[ans_var] = env_var
        store_yaml_file(path, content, nice=True, sort_keys=True)

    def add_integration_tests(
        ansible_core_version: AnsibleCoreVersion,
        repo_name: str | None = None,
        branch_name: str | None = None,
    ) -> list[str]:
        # Determine Python versions to run tests for
        py_versions = (
            (core_python_versions.get(branch_name) if branch_name is not None else None)
            or core_python_versions.get(ansible_core_version)
            or core_python_versions.get(str(ansible_core_version))
            if core_python_versions
            else None
        )
        if py_versions is None:
            core_info = get_ansible_core_info(ansible_core_version)
            py_versions = list(
                core_info.controller_python_versions
                if controller_python_versions_only
                else core_info.remote_python_versions
            )

        # Add sessions
        integration_sessions_core: list[str] = []
        if branch_name is None:
            base_name = f"ansible-test-integration-{ansible_core_version}-"
        else:
            repo_prefix = (
                f"{repo_name.replace('/', '-')}-" if repo_name is not None else ""
            )
            base_name = f"ansible-test-integration-{repo_prefix}{branch_name.replace('/', '-')}-"
        for py_version in py_versions:
            name = f"{base_name}{py_version}"
            if branch_name is None:
                description = (
                    f"Run integration tests from ansible-core {ansible_core_version}'s"
                    f" ansible-test with Python {py_version}"
                )
            else:
                repo_postfix = (
                    f", {repo_name} repository" if repo_name is not None else ""
                )
                description = (
                    f"Run integration tests from ansible-test in ansible-core's {branch_name}"
                    f" branch{repo_postfix} with Python {py_version}"
                )
            add_ansible_test_session(
                name=name,
                description=description,
                ansible_test_params=[
                    "integration",
                    "--color",
                    "-v",
                    "--docker",
                    "default",
                    "--python",
                    str(py_version),
                ],
                extra_deps_files=["tests/integration/requirements.yml"],
                ansible_core_version=ansible_core_version,
                ansible_core_repo_name=repo_name,
                ansible_core_branch_name=branch_name,
                callback_before=callback_before,
                default=False,
                register_name="integration",
                register_extra_data={
                    "display-name": f"Ⓐ{ansible_core_version}+py{py_version}+default"
                },
            )
            integration_sessions_core.append(name)
        return integration_sessions_core

    parsed_min_version, parsed_max_version, parsed_except_versions = (
        _parse_min_max_except(min_version, max_version, except_versions)
    )
    integration_sessions: list[str] = []
    for ansible_core_version in get_supported_core_versions(
        include_devel=include_devel,
        include_milestone=include_milestone,
        min_version=parsed_min_version,
        max_version=parsed_max_version,
        except_versions=parsed_except_versions,
    ):
        integration_sessions_core = add_integration_tests(ansible_core_version)
        if integration_sessions_core:
            name = f"ansible-test-integration-{ansible_core_version}"
            integration_sessions.append(name)

            def run_integration_tests(
                session: nox.Session,  # pylint: disable=unused-argument
            ) -> None:
                pass

            run_integration_tests.__doc__ = (
                f"Meta session for running all {name}-* sessions."
            )
            nox.session(
                name=name,
                requires=integration_sessions_core,
                default=False,
            )(run_integration_tests)
    if add_devel_like_branches:
        for repo_name, branch_name in add_devel_like_branches:
            integration_sessions_core = add_integration_tests(
                "devel", repo_name=repo_name, branch_name=branch_name
            )
            if integration_sessions_core:
                repo_prefix = (
                    f"{repo_name.replace('/', '-')}-" if repo_name is not None else ""
                )
                name = f"ansible-test-integration-{repo_prefix}{branch_name.replace('/', '-')}"
                integration_sessions.append(name)

                def run_integration_tests_for_branch(
                    session: nox.Session,  # pylint: disable=unused-argument
                ) -> None:
                    pass

                run_integration_tests_for_branch.__doc__ = (
                    f"Meta session for running all {name}-* sessions."
                )
                nox.session(
                    name=name,
                    requires=integration_sessions_core,
                    default=False,
                )(run_integration_tests_for_branch)

    def ansible_test_integration(
        session: nox.Session,  # pylint: disable=unused-argument
    ) -> None:
        pass

    ansible_test_integration.__doc__ = (
        "Meta session for running all ansible-test-integration-* sessions."
    )
    nox.session(
        name="ansible-test-integration",
        requires=integration_sessions,
        default=default,
    )(ansible_test_integration)


__all__ = [
    "add_ansible_test_session",
    "add_ansible_test_sanity_test_session",
    "add_all_ansible_test_sanity_test_sessions",
    "add_ansible_test_unit_test_session",
    "add_all_ansible_test_unit_test_sessions",
    "add_ansible_test_integration_sessions_default_container",
]



#!/usr/bin/env python

# Copyright (c) Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt
# or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""Prevent unwanted files from being added to the source tree."""

from __future__ import annotations

import os
import sys

from antsibull_nox.data.antsibull_nox_data_util import (
    get_bool,
    get_list_of_strings,
    setup,
)


def main() -> int:
    """Main entry point."""
    paths, extra_data = setup()

    module_extensions = tuple(
        sorted(
            get_list_of_strings(
                extra_data,
                "module_extensions",
                default=[
                    ".cs",
                    ".ps1",
                    ".psm1",
                    ".py",
                ],
            )
        )
    )

    other_extensions = tuple(
        sorted(
            get_list_of_strings(
                extra_data,
                "other_extensions",
                default=[
                    ".py",
                    ".pyi",
                ],
            )
        )
    )

    yaml_extensions = set(
        get_list_of_strings(
            extra_data,
            "yaml_extensions",
            default=[
                ".yml",
                ".yaml",
            ],
        )
    )

    skip_paths = set(get_list_of_strings(extra_data, "skip_paths", default=[]))

    skip_directories = tuple(
        get_list_of_strings(extra_data, "skip_directories", default=[])
    )

    yaml_directories = tuple(
        get_list_of_strings(
            extra_data,
            "yaml_directories",
            default=[
                "plugins/test/",
                "plugins/filter/",
            ],
        )
    )

    allow_symlinks = get_bool(extra_data, "allow_symlinks")

    errors: list[str] = []
    for path in paths:
        if not path.startswith("plugins/"):
            continue

        if path in skip_paths:
            continue

        if any(path.startswith(skip_directory) for skip_directory in skip_directories):
            continue

        if os.path.islink(path):
            if not allow_symlinks:
                errors.append("%s: is a symbolic link" % (path,))
        elif not os.path.isfile(path):
            errors.append("%s: is not a regular file" % (path,))

        ext = os.path.splitext(path)[1]

        if ext in yaml_extensions and any(
            path.startswith(yaml_directory) for yaml_directory in yaml_directories
        ):
            continue

        extensions = (
            module_extensions
            if path.startswith("plugins/modules/")
            else other_extensions
        )

        if ext not in extensions:
            errors.append(f"{path}: extension must be one of: {', '.join(extensions)}")

    for error in sorted(errors):
        print(error)
    return len(errors) > 0


if __name__ == "__main__":
    sys.exit(main())



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

from __future__ import annotations

import contextlib
import os
import typing as t
from pathlib import Path

from antsibull_nox.python import get_installed_python_versions
from antsibull_nox.utils import Version


@contextlib.contextmanager
def set_environ(env_var: str, value: str) -> t.Iterator[None]:
    old_value = os.environ.get(env_var)
    try:
        os.environ[env_var] = value
        yield
    finally:
        if old_value is None:
            os.environ.pop(env_var, None)
        else:
            os.environ[env_var] = old_value


def fake_binary(path: Path, content: str) -> Path:
    path.write_text(content)
    path.chmod(0o700)
    return path


def fake_python_binary(path: Path, version: str) -> Path:
    return fake_binary(path, f"""#!/bin/bash\necho {version}\n""")


def test_get_installed_python_versions(tmp_path: Path) -> None:
    bin_dir = tmp_path / "bin"
    bin_dir.mkdir()
    with set_environ("PATH", str(bin_dir)):
        get_installed_python_versions.cache_clear()  # added by @functools.cache
        assert get_installed_python_versions() == {}

    py27 = fake_python_binary(bin_dir / "python2.7", "2.7")
    with set_environ("PATH", str(bin_dir)):
        get_installed_python_versions.cache_clear()  # added by @functools.cache
        assert get_installed_python_versions() == {Version(2, 7): py27}

    py27 = fake_python_binary(bin_dir / "python2.7", "2.7")
    py35 = fake_python_binary(bin_dir / "python3", "3.5")
    fake_binary(bin_dir / "python2", "#!/bin/bash\necho Foo\n")
    fake_binary(bin_dir / "python", "#!/bin/bash\nexit 1\n")
    with set_environ("PATH", str(bin_dir)):
        get_installed_python_versions.cache_clear()  # added by @functools.cache
        assert get_installed_python_versions() == {
            Version(2, 7): py27,
            Version(3, 5): py35,
        }



#!/usr/bin/env python

# Copyright (c) 2022, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt
# or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""Prevent files without a correct license identifier from being added to the source tree."""

from __future__ import annotations

import glob
import os
import stat
import sys

from antsibull_nox.data.antsibull_nox_data_util import get_list_of_strings, setup


def format_license_list(licenses: list[str]) -> str:
    if not licenses:
        return "(empty)"
    return ", ".join([f'"{license}"' for license in licenses])


def find_licenses(errors: list[str], filename: str, relax: bool = False) -> list[str]:
    spdx_license_identifiers: list[str] = []
    other_license_identifiers: list[str] = []
    has_copyright = False
    try:
        with open(filename, "r", encoding="utf-8") as f:
            for line in f:
                line = line.rstrip()
                if "Copyright " in line:
                    has_copyright = True
                if "Copyright: " in line:
                    errors.append(
                        f'{filename}: found copyright line with "Copyright:".'
                        " Please remove the colon."
                    )
                if "SPDX-FileCopyrightText: " in line:
                    has_copyright = True
                idx = line.find("SPDX-License-Identifier: ")
                if idx >= 0:
                    lic_id = line[idx + len("SPDX-License-Identifier: ") :]
                    spdx_license_identifiers.extend(lic_id.split(" OR "))
                if "GNU General Public License" in line:
                    if "v3.0+" in line:
                        other_license_identifiers.append("GPL-3.0-or-later")
                    if "version 3 or later" in line:
                        other_license_identifiers.append("GPL-3.0-or-later")
                if "Simplified BSD License" in line:
                    other_license_identifiers.append("BSD-2-Clause")
                if "Apache License 2.0" in line:
                    other_license_identifiers.append("Apache-2.0")
                if "PSF License" in line or "Python-2.0" in line:
                    other_license_identifiers.append("PSF-2.0")
                if "MIT License" in line:
                    other_license_identifiers.append("MIT")
    except Exception as exc:
        errors.append(f"{filename}: error while processing file: {exc}")
    if len(set(spdx_license_identifiers)) < len(spdx_license_identifiers):
        errors.append(f"{filename}: found identical SPDX-License-Identifier values")
    if other_license_identifiers and set(other_license_identifiers) != set(
        spdx_license_identifiers
    ):
        errors.append(
            f"{filename}: SPDX-License-Identifier yielded the license list"
            f" {format_license_list(spdx_license_identifiers)}, while manual guessing"
            f" yielded the license list {format_license_list(other_license_identifiers)}"
        )
    if not has_copyright and not relax:
        errors.append(f"{filename}: found no copyright notice")
    return sorted(spdx_license_identifiers)


def main() -> int:
    """Main entry point."""
    paths, extra_data = setup()

    # The following paths are allowed to have no license identifier
    no_comments_allowed = [
        "changelogs/fragments/*.yml",
        "changelogs/fragments/*.yaml",
    ]

    # These files are completely ignored
    ignore_paths = [
        ".ansible-test-timeout.json",
        ".reuse/dep5",
        "LICENSES/*.txt",
        "COPYING",
    ] + get_list_of_strings(extra_data, "extra_ignore_paths", default=[])

    no_comments_allowed = [
        fn for pattern in no_comments_allowed for fn in glob.glob(pattern)
    ]
    ignore_paths = [fn for pattern in ignore_paths for fn in glob.glob(pattern)]

    valid_licenses = [
        license_file[len("LICENSES/") : -len(".txt")]
        for license_file in glob.glob("LICENSES/*.txt")
    ]

    errors: list[str] = []

    for path in paths:
        if path.startswith("./"):
            path = path[2:]
        if path in ignore_paths or path.startswith("tests/output/"):
            continue
        sr = os.stat(path)
        if not stat.S_ISREG(sr.st_mode):
            continue
        if sr.st_size == 0:
            continue
        if not path.endswith(".license") and os.path.exists(path + ".license"):
            path = path + ".license"
        valid_licenses_for_path = valid_licenses
        if (
            path.startswith("plugins/")
            and not path.startswith(
                ("plugins/modules/", "plugins/module_utils/", "plugins/doc_fragments/")
            )
            and path.endswith((".py", ".py.license"))
        ):
            valid_licenses_for_path = [
                license for license in valid_licenses if license == "GPL-3.0-or-later"
            ]
        licenses = find_licenses(errors, path, relax=path in no_comments_allowed)
        if not licenses:
            if path not in no_comments_allowed:
                errors.append(f"{path}: must have at least one license")
        else:
            for license in licenses:
                if license not in valid_licenses_for_path:
                    errors.append(
                        f"{path}: found not allowed license {license!r}, must be one of"
                        f" {format_license_list(valid_licenses_for_path)}"
                    )

    for error in sorted(errors):
        print(error)
    return len(errors) > 0


if __name__ == "__main__":
    sys.exit(main())



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Config file schema.
"""

from __future__ import annotations

import os
import typing as t

import pydantic as p

from antsibull_nox.ee_config import create_ee_config

from ._pydantic import forbid_extras, get_formatted_error_messages
from .ansible import AnsibleCoreVersion
from .utils import Version

try:
    from tomllib import load as _load_toml
except ImportError:
    from tomli import load as _load_toml  # type: ignore


CONFIG_FILENAME = "antsibull-nox.toml"


def _parse_version(value: t.Any) -> Version:
    if isinstance(value, Version):
        return value
    if isinstance(value, str) and "." in value:
        return Version.parse(value)
    raise ValueError("Must be version string")


def _parse_ansible_core_version(value: t.Any) -> AnsibleCoreVersion:
    if isinstance(value, Version):
        return value
    if isinstance(value, str):
        if value == "devel":
            return "devel"
        if value == "milestone":
            return "milestone"
        if "." in value:
            return Version.parse(value)
    raise ValueError("Must be ansible-core version string")


def _validate_collection_name(value: str) -> str:
    parts = value.split(".")
    if len(parts) != 2:
        raise ValueError("Collection name must be of the form '<namespace>.<name>'")
    if not parts[0].isidentifier():
        raise ValueError("Collection namespace must be Python identifier")
    if not parts[1].isidentifier():
        raise ValueError("Collection name must be Python identifier")
    return value


CollectionName = t.Annotated[str, p.AfterValidator(_validate_collection_name)]
PVersion = t.Annotated[Version, p.BeforeValidator(_parse_version)]
PAnsibleCoreVersion = t.Annotated[
    AnsibleCoreVersion, p.BeforeValidator(_parse_ansible_core_version)
]


class _BaseModel(p.BaseModel):
    model_config = p.ConfigDict(frozen=True, extra="allow", validate_default=True)


class SessionLint(_BaseModel):
    """
    Lint session config.
    """

    default: bool = True
    extra_code_files: list[str] = []
    ruff_config: t.Optional[p.FilePath] = None
    ruff_package: str = "ruff"

    # isort:
    run_isort: bool = True
    isort_config: t.Optional[p.FilePath] = None
    isort_package: str = "isort"

    # black:
    run_black: bool = True
    run_black_modules: t.Optional[bool] = None
    black_config: t.Optional[p.FilePath] = None
    black_package: str = "black"

    # ruff format:
    run_ruff_format: bool = False
    ruff_format_config: t.Optional[p.FilePath] = None
    ruff_format_package: t.Optional[str] = None

    # ruff autofix:
    run_ruff_autofix: bool = False
    ruff_autofix_config: t.Optional[p.FilePath] = None
    ruff_autofix_package: t.Optional[str] = None
    ruff_autofix_select: list[str] = []

    # ruff check:
    run_ruff_check: bool = False
    ruff_check_config: t.Optional[p.FilePath] = None
    ruff_check_package: t.Optional[str] = None

    # flake8:
    run_flake8: bool = True
    flake8_config: t.Optional[p.FilePath] = None
    flake8_package: str = "flake8"

    # pylint:
    run_pylint: bool = True
    pylint_rcfile: t.Optional[p.FilePath] = None
    pylint_modules_rcfile: t.Optional[p.FilePath] = None
    pylint_package: str = "pylint"
    pylint_ansible_core_package: t.Optional[str] = "ansible-core"
    pylint_extra_deps: list[str] = []

    # yamllint:
    run_yamllint: bool = True
    yamllint_config: t.Optional[p.FilePath] = None
    yamllint_config_plugins: t.Optional[p.FilePath] = None
    yamllint_config_plugins_examples: t.Optional[p.FilePath] = None
    yamllint_config_extra_docs: t.Optional[p.FilePath] = None
    yamllint_package: str = "yamllint"
    yamllint_antsibull_docutils_package: str = "antsibull-docutils"

    # mypy:
    run_mypy: bool = True
    mypy_config: t.Optional[p.FilePath] = None
    mypy_package: str = "mypy"
    mypy_ansible_core_package: t.Optional[str] = "ansible-core"
    mypy_extra_deps: list[str] = []

    # antsibull-nox config lint:
    run_antsibullnox_config_lint: bool = True


class SessionDocsCheck(_BaseModel):
    """
    Docs check session config.
    """

    default: bool = True

    antsibull_docs_package: str = "antsibull-docs"
    ansible_core_package: str = "ansible-core"
    validate_collection_refs: t.Optional[t.Literal["self", "dependent", "all"]] = None
    extra_collections: list[CollectionName] = []

    codeblocks_restrict_types: t.Optional[list[str]] = None
    codeblocks_restrict_type_exact_case: bool = True
    codeblocks_allow_without_type: bool = True
    codeblocks_allow_literal_blocks: bool = True
    antsibull_docutils_package: str = "antsibull-docutils"


class SessionLicenseCheck(_BaseModel):
    """
    License check session config.
    """

    default: bool = True

    run_reuse: bool = True
    reuse_package: str = "reuse"
    run_license_check: bool = True
    license_check_extra_ignore_paths: list[str] = []


class ExecutionEnvironmentConfig(_BaseModel):
    """
    Execution enviroment check session config.
    """

    name: str
    description: t.Optional[str] = None

    # EE definition
    version: t.Literal[3] = 3
    base_image_name: t.Optional[str] = None  # implicit default
    ansible_core_source: t.Literal["package_pip", "package_system"] = "package_pip"
    ansible_core_package: t.Optional[str] = None
    ansible_runner_source: t.Literal["package_pip", "package_system"] = "package_pip"
    ansible_runner_package: t.Optional[str] = None
    system_packages: list[str] = []
    python_packages: list[str] = []
    python_interpreter_package: t.Optional[str] = None
    python_path: t.Optional[str] = None
    config: dict[str, t.Any] = {}

    # EE tests
    test_playbooks: list[str]
    runtime_environment: dict[str, str] = {}
    runtime_container_options: list[str] = []
    runtime_extra_vars: dict[str, str] = {}

    def to_execution_environment_config(self) -> dict[str, t.Any]:
        """
        Convert TOML config to execution environment YAML.
        """

        dependencies: dict[str, t.Any] = {}
        if self.ansible_core_package is not None:
            dependencies["ansible_core"] = {
                self.ansible_core_source: self.ansible_core_package
            }
        if self.ansible_runner_package is not None:
            dependencies["ansible_runner"] = {
                self.ansible_runner_source: self.ansible_runner_package
            }
        if self.python_interpreter_package is not None:
            python_interpreter: dict[str, t.Any] = {
                "package_system": self.python_interpreter_package
            }
            if self.python_path is not None:
                python_interpreter["python_path"] = self.python_path
            dependencies["python_interpreter"] = python_interpreter
        if self.system_packages:
            dependencies["system"] = self.system_packages
        if self.python_packages:
            dependencies["python"] = self.python_packages

        simple_config = create_ee_config(
            version=self.version,
            base_image=(
                "registry.fedoraproject.org/fedora-toolbox:latest"
                if self.base_image_name is None
                else self.base_image_name
            ),
            base_image_is_default=self.base_image_name is None,
            dependencies=dependencies,
            config=self.config,
        )

        ee_config = {**simple_config, **self.config}

        return ee_config


class SessionExecutionEnvironmentCheck(_BaseModel):
    """
    Execution environment check session.
    """

    default: bool = False
    ansible_builder_package: str = "ansible-builder"
    ansible_core_package: t.Optional[str] = None
    ansible_navigator_package: str = "ansible-navigator"

    execution_environments: list[ExecutionEnvironmentConfig]


class ActionGroup(_BaseModel):
    """
    Information about an action group.
    """

    # Name of the action group.
    name: str
    # Regex pattern to match modules that could belong to this action group.
    pattern: str
    # Doc fragment that members of the action group must have, but no other module
    # must have
    doc_fragment: str
    # Exclusion list of modules that match the regex, but should not be part of the
    # action group. All other modules matching the regex are assumed to be part of
    # the action group.
    exclusions: list[str] = []


class AvoidCharacterGroup(_BaseModel):
    """
    Information about characters/regexes to avoid in files.
    """

    # User-friendly name
    name: t.Optional[str] = None

    # Regular expression
    regex: str

    # Extensions, paths, and directories to look for.
    # If None (not specified), will consider all files.
    match_extensions: t.Optional[list[str]] = None
    match_paths: t.Optional[list[str]] = None
    match_directories: t.Optional[list[str]] = None

    # Extensions, paths, and directories to skip.
    skip_extensions: list[str] = []
    skip_paths: list[str] = []
    skip_directories: list[str] = []


class SessionExtraChecks(_BaseModel):
    """
    Extra checks session config.
    """

    default: bool = True

    # no-unwanted-files:
    run_no_unwanted_files: bool = True
    no_unwanted_files_module_extensions: list[str] = [".cs", ".ps1", ".psm1", ".py"]
    no_unwanted_files_other_extensions: list[str] = [".py", ".pyi"]
    no_unwanted_files_yaml_extensions: list[str] = [".yml", ".yaml"]
    no_unwanted_files_skip_paths: list[str] = []
    no_unwanted_files_skip_directories: list[str] = []
    no_unwanted_files_yaml_directories: list[str] = [
        "plugins/test/",
        "plugins/filter/",
    ]
    no_unwanted_files_allow_symlinks: bool = False

    # action-groups:
    run_action_groups: bool = False
    action_groups_config: list[ActionGroup] = []

    # no-trailing-whitespace:
    run_no_trailing_whitespace: bool = False
    no_trailing_whitespace_skip_paths: list[str] = []
    no_trailing_whitespace_skip_directories: list[str] = []

    # avoid-characters:
    run_avoid_characters: bool = False
    avoid_character_group: list[AvoidCharacterGroup] = []


class SessionBuildImportCheck(_BaseModel):
    """
    Collection build and Galaxy import session config.
    """

    default: bool = True

    ansible_core_package: str = "ansible-core"
    run_galaxy_importer: bool = True
    galaxy_importer_package: str = "galaxy-importer"
    # https://github.com/ansible/galaxy-importer#configuration
    galaxy_importer_config_path: t.Optional[p.FilePath] = None
    galaxy_importer_always_show_logs: bool = False


class DevelLikeBranch(_BaseModel):
    """
    A Git repository + branch for a devel-like branch of ansible-core.
    """

    repository: t.Optional[str] = None
    branch: str

    @p.model_validator(mode="before")
    @classmethod
    def _pre_validate(cls, values: t.Any) -> t.Any:
        if isinstance(values, str):
            return {"branch": values}
        if (
            isinstance(values, list)
            and len(values) == 2
            and all(isinstance(v, str) for v in values)
        ):
            return {"repository": values[0], "branch": values[1]}
        return values


class SessionAnsibleTestSanity(_BaseModel):
    """
    Ansible-test sanity tests session config.
    """

    default: bool = False

    include_devel: bool = False
    include_milestone: bool = False
    add_devel_like_branches: list[DevelLikeBranch] = []
    min_version: t.Optional[PVersion] = None
    max_version: t.Optional[PVersion] = None
    except_versions: list[PAnsibleCoreVersion] = []
    skip_tests: list[str] = []
    allow_disabled: bool = False
    enable_optional_errors: bool = False


class SessionAnsibleTestUnits(_BaseModel):
    """
    Ansible-test unit tests session config.
    """

    default: bool = False

    include_devel: bool = False
    include_milestone: bool = False
    add_devel_like_branches: list[DevelLikeBranch] = []
    min_version: t.Optional[PVersion] = None
    max_version: t.Optional[PVersion] = None
    except_versions: list[PAnsibleCoreVersion] = []


class SessionAnsibleTestIntegrationWDefaultContainer(_BaseModel):
    """
    Ansible-test integration tests with default container session config.
    """

    default: bool = False

    include_devel: bool = False
    include_milestone: bool = False
    add_devel_like_branches: list[DevelLikeBranch] = []
    min_version: t.Optional[PVersion] = None
    max_version: t.Optional[PVersion] = None
    except_versions: list[PAnsibleCoreVersion] = []
    core_python_versions: dict[t.Union[PAnsibleCoreVersion, str], list[PVersion]] = {}
    controller_python_versions_only: bool = False
    ansible_vars_from_env_vars: dict[str, str] = {}

    @p.model_validator(mode="after")
    def _validate_core_keys(self) -> t.Self:
        branch_names = [dlb.branch for dlb in self.add_devel_like_branches]
        for key in self.core_python_versions:
            if isinstance(key, Version) or key in {"devel", "milestone"}:
                continue
            if key in branch_names:
                continue
            raise ValueError(
                f"Unknown ansible-core version or branch name {key!r} in core_python_versions"
            )
        return self


class SessionAnsibleLint(_BaseModel):
    """
    Ansible-lint session config.
    """

    default: bool = True

    ansible_lint_package: str = "ansible-lint"
    strict: bool = False


class Sessions(_BaseModel):
    """
    Configuration of nox sessions to add.
    """

    lint: t.Optional[SessionLint] = None
    docs_check: t.Optional[SessionDocsCheck] = None
    license_check: t.Optional[SessionLicenseCheck] = None
    extra_checks: t.Optional[SessionExtraChecks] = None
    build_import_check: t.Optional[SessionBuildImportCheck] = None
    ansible_test_sanity: t.Optional[SessionAnsibleTestSanity] = None
    ansible_test_units: t.Optional[SessionAnsibleTestUnits] = None
    ansible_test_integration_w_default_container: t.Optional[
        SessionAnsibleTestIntegrationWDefaultContainer
    ] = None
    ansible_lint: t.Optional[SessionAnsibleLint] = None
    ee_check: t.Optional[SessionExecutionEnvironmentCheck] = None


class CollectionSource(_BaseModel):
    """
    Source from which to install a collection.
    """

    source: str

    @p.model_validator(mode="before")
    @classmethod
    def _pre_validate(cls, values):
        if isinstance(values, str):
            return {"source": values}
        return values


class Config(_BaseModel):
    """
    The contents of a antsibull-nox config file.
    """

    collection_sources: dict[CollectionName, CollectionSource] = {}
    collection_sources_per_ansible: dict[
        PAnsibleCoreVersion, dict[CollectionName, CollectionSource]
    ] = {}
    sessions: Sessions = Sessions()


def load_config_from_toml(path: str | os.PathLike) -> Config:
    """
    Load a config TOML file.
    """
    with open(path, "rb") as f:
        try:
            data = _load_toml(f)
        except ValueError as exc:
            raise ValueError(f"Error while reading {path}: {exc}") from exc
    return Config.model_validate(data)


def lint_config_toml() -> list[str]:
    """
    Lint config files
    """
    path = CONFIG_FILENAME
    errors = []
    forbid_extras(Config)
    try:
        with open(path, "rb") as f:
            data = _load_toml(f)
        Config.model_validate(data)
    except p.ValidationError as exc:
        for error in get_formatted_error_messages(exc):
            errors.append(f"{path}:{error}")
    except ValueError as exc:
        errors.append(f"{path}:{exc}")
    except FileNotFoundError:
        errors.append(f"{path}: File does not exist")
    except IOError as exc:
        errors.append(f"{path}:{exc}")
    return errors



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2024

# ========================================================================
# **Vendored** from antsibull-core.
#
# I vendored this to avoid depending on antsibull-core only for this.
# TBH, most of this should be part of pydantic anyway. See
# https://github.com/pydantic/pydantic/discussions/2652 for a discussion
# on this.
#
# pylint: disable=missing-function-docstring
# ========================================================================

"""
Helpers for pydantic.
"""

from __future__ import annotations

import typing as t
from collections.abc import Callable, Collection

import pydantic as p

if t.TYPE_CHECKING:
    from typing_extensions import TypeGuard


def _is_basemodel(a_type: t.Any) -> TypeGuard[type[p.BaseModel]]:
    try:
        return issubclass(a_type, p.BaseModel)
    except TypeError:
        # If inspect.isclass(a_type) is checked first, no TypeError happens for
        # Python 3.11+.

        # On Python 3.9 and 3.10, issubclass(dict[int, int], p.BaseModel) raises
        # "TypeError: issubclass() arg 1 must be a class".
        # (https://github.com/pydantic/pydantic/discussions/5970)
        return False


def _modify_config(
    cls: type[p.BaseModel],
    processed_classes: set[type[p.BaseModel]],
    change_config: Callable[[p.ConfigDict], bool],
) -> bool:
    if cls in processed_classes:
        return False
    change = False
    for field_info in cls.model_fields.values():
        if _is_basemodel(field_info.annotation):
            change |= _modify_config(
                field_info.annotation,
                processed_classes,
                change_config,
            )
        for subcls in t.get_args(field_info.annotation):
            if _is_basemodel(subcls):
                change |= _modify_config(subcls, processed_classes, change_config)
    change |= change_config(cls.model_config)
    if change:
        cls.model_rebuild(force=True)
    processed_classes.add(cls)
    return change


def set_extras(
    models: type[p.BaseModel] | Collection[type[p.BaseModel]],
    value: t.Literal["allow", "ignore", "forbid"],
) -> None:
    def change_config(model_config: p.ConfigDict) -> bool:
        if model_config.get("extra") == value:
            return False
        model_config["extra"] = value
        return True

    processed_classes: set[type[p.BaseModel]] = set()
    if isinstance(models, Collection):
        for cls in models:
            _modify_config(cls, processed_classes, change_config)
    else:
        _modify_config(models, processed_classes, change_config)


def forbid_extras(models: type[p.BaseModel] | Collection[type[p.BaseModel]]) -> None:
    set_extras(models, "forbid")


def get_formatted_error_messages(error: p.ValidationError) -> list[str]:
    def format_error(err) -> str:
        location = " -> ".join(str(loc) for loc in err["loc"])
        return f'{location}: {err["msg"]}'

    return [format_error(err) for err in error.errors()]



#!/usr/bin/env python

# Copyright (c) 2025, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt
# or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""Make sure all plugin and module documentation adheres to yamllint."""

from __future__ import annotations

import ast
import io
import os
import re
import sys
import traceback
import typing as t

import yaml
from antsibull_nox_data_util import setup  # type: ignore
from yamllint import linter
from yamllint.config import YamlLintConfig
from yamllint.linter import PROBLEM_LEVELS

REPORT_LEVELS: set[PROBLEM_LEVELS] = {
    "warning",
    "error",
}

EXAMPLES_FMT_RE = re.compile(r"^# fmt:\s+(\S+)")

EXAMPLES_SECTION = "EXAMPLES"


def lint(
    *,
    errors: list[dict[str, t.Any]],
    path: str,
    data: str,
    row_offset: int,
    col_offset: int,
    section: str,
    config: YamlLintConfig,
    extra_for_errors: dict[str, t.Any] | None = None,
) -> None:
    # If the string start with optional whitespace + linebreak, skip that line
    idx = data.find("\n")
    if idx >= 0 and (idx == 0 or data[:idx].isspace()):
        data = data[idx + 1 :]
        row_offset += 1
        col_offset = 0

    try:
        problems = linter.run(
            io.StringIO(data),
            config,
            path,
        )
        for problem in problems:
            if problem.level not in REPORT_LEVELS:
                continue
            msg = f"{section}: {problem.level}: {problem.desc}"
            if problem.rule:
                msg += f"  ({problem.rule})"
            errors.append(
                {
                    "path": path,
                    "line": row_offset + problem.line,
                    # The col_offset is only valid for line 1; otherwise the offset is 0
                    "col": (col_offset if problem.line == 1 else 0) + problem.column,
                    "message": msg,
                }
            )
            if extra_for_errors:
                errors[-1].update(extra_for_errors)
    except Exception as exc:
        error = str(exc).replace("\n", " / ")
        errors.append(
            {
                "path": path,
                "line": row_offset + 1,
                "col": col_offset + 1,
                "message": (
                    f"{section}: Internal error while linting YAML: exception {type(exc)}:"
                    f" {error}; traceback: {traceback.format_exc()!r}"
                ),
            }
        )
        if extra_for_errors:
            errors[-1].update(extra_for_errors)


def iterate_targets(
    assignment: ast.Assign,
) -> t.Iterable[tuple[ast.Constant, str, str]]:
    if not isinstance(assignment.value, ast.Constant):
        return
    if not isinstance(assignment.value.value, str):
        return
    for target in assignment.targets:
        try:
            yield assignment.value, assignment.value.value, target.id  # type: ignore
        except AttributeError:
            continue


def process_python_file(
    errors: list[dict[str, t.Any]],
    path: str,
    config: YamlLintConfig,
    config_examples: YamlLintConfig,
) -> None:
    try:
        with open(path, "rt", encoding="utf-8") as f:
            root = ast.parse(f.read(), filename=path)
    except Exception as exc:
        errors.append(
            {
                "path": path,
                "line": 1,
                "col": 1,
                "message": (
                    f"Error while parsing Python code: exception {type(exc)}:"
                    f" {exc}; traceback: {traceback.format_exc()!r}"
                ),
            }
        )
        return

    is_doc_fragment = path.startswith("plugins/doc_fragments/")

    # We look for top-level assignments and classes
    for child in root.body:
        if (
            is_doc_fragment
            and isinstance(child, ast.ClassDef)
            and child.name == "ModuleDocFragment"
        ):
            for fragment in child.body:
                if not isinstance(fragment, ast.Assign):
                    continue
                for constant, data, fragment_name in iterate_targets(fragment):
                    lint(
                        errors=errors,
                        path=path,
                        data=data,
                        row_offset=constant.lineno - 1,
                        col_offset=constant.col_offset - 1,
                        section=fragment_name,
                        config=config,
                    )
        if not isinstance(child, ast.Assign):
            continue
        for constant, data, section in iterate_targets(child):
            if section not in ("DOCUMENTATION", "EXAMPLES", "RETURN"):
                continue

            # Handle special values
            if data in ("#", " # ") and section == "RETURN":
                # Not skipping it here could result in all kind of linting errors,
                # like no document start, or trailing space.
                continue

            # Check for non-YAML examples
            if section == EXAMPLES_SECTION:
                fmt_match = EXAMPLES_FMT_RE.match(data.lstrip())
                if fmt_match and fmt_match.group(1) != "yaml":
                    continue

            # Parse the (remaining) string content
            lint(
                errors=errors,
                path=path,
                data=data,
                row_offset=constant.lineno - 1,
                col_offset=constant.col_offset - 1,
                section=section,
                config=config_examples if section == EXAMPLES_SECTION else config,
            )


def process_sidecar_docs_file(
    errors: list[dict[str, t.Any]],
    path: str,
    config_examples: YamlLintConfig,
) -> None:
    try:
        # TODO: get hold of YAML structure so we also get correct line/col numbers
        #       inside EXAMPLES!
        with open(path, "rb") as stream:
            root = yaml.load(stream, Loader=yaml.SafeLoader)
    except Exception as exc:
        errors.append(
            {
                "path": path,
                "line": 1,
                "col": 1,
                "message": (
                    f"Error while parsing Python code: exception {type(exc)}:"
                    f" {exc}; traceback: {traceback.format_exc()!r}"
                ),
            }
        )
        return

    if not isinstance(root, dict):
        return
    examples = root.get(EXAMPLES_SECTION)
    if not isinstance(examples, str):
        return

    # Check for non-YAML examples
    fmt_match = EXAMPLES_FMT_RE.match(examples.lstrip())
    if fmt_match and fmt_match.group(1) != "yaml":
        return

    lint(
        errors=errors,
        path=path,
        data=examples,
        row_offset=0,  # TODO
        col_offset=0,  # TODO
        section=EXAMPLES_SECTION,
        config=config_examples,
        extra_for_errors={
            "note": "Line/column are relative to EXAMPLES string contents"
        },
    )


def main() -> int:
    """Main entry point."""
    paths, extra_data = setup()
    config: str | None = extra_data.get("config")
    config_examples: str | None = extra_data.get("config_examples")

    if config:
        yamllint_config = YamlLintConfig(file=config)
    else:
        yamllint_config = YamlLintConfig(content="extends: default")

    if config_examples:
        yamllint_config_examples = YamlLintConfig(file=config_examples)
    else:
        yamllint_config_examples = yamllint_config

    errors: list[dict[str, t.Any]] = []
    for path in paths:
        if not os.path.isfile(path):
            continue
        if path.endswith(".py"):
            process_python_file(errors, path, yamllint_config, yamllint_config_examples)
        if path.endswith((".yml", ".yaml")):
            process_sidecar_docs_file(errors, path, yamllint_config_examples)

    errors.sort(
        key=lambda error: (error["path"], error["line"], error["col"], error["message"])
    )
    for error in errors:
        prefix = f"{error['path']}:{error['line']}:{error['col']}: "
        msg = error["message"]
        if "note" in error:
            msg = f"{msg}\nNote: {error['note']}"
        for i, line in enumerate(msg.splitlines()):
            print(f"{prefix}{line}")
            if i == 0:
                prefix = " " * len(prefix)

    return len(errors) > 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

# Copyright (c) 2025, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt
# or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""Make sure all YAML in RST extra documentation adheres to yamllint."""

from __future__ import annotations

import io
import os
import sys
import traceback
import typing as t

from antsibull_docutils.rst_code_finder import (
    find_code_blocks,
    mark_antsibull_code_block,
)
from antsibull_nox_data_util import setup  # type: ignore
from docutils import nodes
from docutils.parsers.rst import Directive
from yamllint import linter
from yamllint.config import YamlLintConfig
from yamllint.linter import PROBLEM_LEVELS

REPORT_LEVELS: set[PROBLEM_LEVELS] = {
    "warning",
    "error",
}

YAML_LANGUAGES = {"yaml", "yaml+jinja"}


def lint(
    *,
    errors: list[dict[str, t.Any]],
    path: str,
    data: str,
    row_offset: int,
    col_offset: int,
    config: YamlLintConfig,
    extra_for_errors: dict[str, t.Any] | None = None,
) -> None:
    # If the string start with optional whitespace + linebreak, skip that line
    idx = data.find("\n")
    if idx >= 0 and (idx == 0 or data[:idx].isspace()):
        data = data[idx + 1 :]
        row_offset += 1
        col_offset = 0

    try:
        problems = linter.run(
            io.StringIO(data),
            config,
            path,
        )
        for problem in problems:
            if problem.level not in REPORT_LEVELS:
                continue
            msg = f"{problem.level}: {problem.desc}"
            if problem.rule:
                msg += f"  ({problem.rule})"
            errors.append(
                {
                    "path": path,
                    "line": row_offset + problem.line,
                    # The col_offset is only valid for line 1; otherwise the offset is 0
                    "col": (col_offset if problem.line == 1 else 0) + problem.column,
                    "message": msg,
                }
            )
            if extra_for_errors:
                errors[-1].update(extra_for_errors)
    except Exception as exc:
        error = str(exc).replace("\n", " / ")
        errors.append(
            {
                "path": path,
                "line": row_offset + 1,
                "col": col_offset + 1,
                "message": (
                    f"Internal error while linting YAML: exception {type(exc)}:"
                    f" {error}; traceback: {traceback.format_exc()!r}"
                ),
            }
        )
        if extra_for_errors:
            errors[-1].update(extra_for_errors)


_ANSIBLE_OUTPUT_DATA_LANGUAGE = "ansible-output-data-FPho6oogookao7okinoX"
_ANSIBLE_OUTPUT_META_LANGUAGE = "ansible-output-meta-FPho6oogookao7okinoX"


class AnsibleOutputDataDirective(Directive):
    has_content = True

    def run(self) -> list[nodes.literal_block]:
        code = "\n".join(self.content)
        literal = nodes.literal_block(code, code)
        literal["classes"].append("code-block")
        mark_antsibull_code_block(
            literal,
            language=_ANSIBLE_OUTPUT_DATA_LANGUAGE,
            content_offset=self.content_offset,
        )
        return [literal]


class AnsibleOutputMetaDirective(Directive):
    has_content = True

    def run(self) -> list[nodes.literal_block]:
        code = "\n".join(self.content)
        literal = nodes.literal_block(code, code)
        literal["classes"].append("code-block")
        mark_antsibull_code_block(
            literal,
            language=_ANSIBLE_OUTPUT_META_LANGUAGE,
            content_offset=self.content_offset,
        )
        return [literal]


def process_rst_file(
    errors: list[dict[str, t.Any]],
    path: str,
    config: YamlLintConfig,
) -> None:
    try:
        with open(path, "rt", encoding="utf-8") as f:
            content = f.read()
    except Exception as exc:
        errors.append(
            {
                "path": path,
                "line": 1,
                "col": 1,
                "message": (
                    f"Error while reading content: {type(exc)}:"
                    f" {exc}; traceback: {traceback.format_exc()!r}"
                ),
            }
        )
        return

    def warn_unknown_block(line: int | str, col: int, content: str) -> None:
        errors.append(
            {
                "path": path,
                "line": line,
                "col": col,
                "message": (
                    "Warning: found unknown literal block! Check for double colons '::'."
                    " If that is not the cause, please report this warning."
                    " It might indicate a bug in the checker or an unsupported Sphinx directive."
                    f" Content: {content!r}"
                ),
            }
        )

    for code_block in find_code_blocks(
        content,
        path=path,
        root_prefix="docs/docsite/rst",
        warn_unknown_block=warn_unknown_block,
        extra_directives={
            "ansible-output-data": AnsibleOutputDataDirective,
            "ansible-output-meta": AnsibleOutputMetaDirective,
        },
    ):
        if (
            code_block.language or ""
        ).lower() not in YAML_LANGUAGES and code_block.language not in (
            _ANSIBLE_OUTPUT_DATA_LANGUAGE,
            _ANSIBLE_OUTPUT_META_LANGUAGE,
        ):
            continue

        extra_for_errors = {}
        if not code_block.position_exact:
            extra_for_errors["note"] = (
                "The code block could not be exactly located in the source file."
                " The line/column numbers might be off."
            )

        # Parse the (remaining) string content
        lint(
            errors=errors,
            path=path,
            data=code_block.content,
            row_offset=code_block.row_offset,
            col_offset=code_block.col_offset,
            config=config,
            extra_for_errors=extra_for_errors,
        )


def main() -> int:
    """Main entry point."""
    paths, extra_data = setup()
    config: str | None = extra_data.get("config")

    if config:
        yamllint_config = YamlLintConfig(file=config)
    else:
        yamllint_config = YamlLintConfig(content="extends: default")

    errors: list[dict[str, t.Any]] = []
    for path in paths:
        if not os.path.isfile(path):
            continue
        process_rst_file(errors, path, yamllint_config)

    errors.sort(
        key=lambda error: (
            error["path"],
            error["line"] if isinstance(error["line"], int) else 0,
            error["col"],
            error["message"],
        )
    )
    for error in errors:
        prefix = f"{error['path']}:{error['line']}:{error['col']}: "
        msg = error["message"]
        if "note" in error:
            msg = f"{msg}\nNote: {error['note']}"
        for i, line in enumerate(msg.splitlines()):
            print(f"{prefix}{line}")
            if i == 0:
                prefix = " " * len(prefix)

    return len(errors) > 0


if __name__ == "__main__":
    sys.exit(main())



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Handle Ansible collections.
"""

from __future__ import annotations

from .build import build_collection
from .data import CollectionData, CollectionSource, SetupResult
from .install import (
    Runner,
    setup_collection_sources,
    setup_collections,
    setup_current_tree,
)
from .search import CollectionList, load_collection_data_from_disk
from .utils import force_collection_version

__all__ = [
    "CollectionData",
    "CollectionList",
    "CollectionSource",
    "SetupResult",
    "Runner",
    "build_collection",
    "force_collection_version",
    "load_collection_data_from_disk",
    "setup_collections",
    "setup_current_tree",
    "setup_collection_sources",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Create nox build import check session.
"""

from __future__ import annotations

import os
import sys
from pathlib import Path

import nox

from ..collection import (
    build_collection,
)
from .utils import (
    ci_group,
    compose_description,
    install,
    nox_has_verbosity,
    silence_run_verbosity,
)


def add_build_import_check(
    *,
    make_build_import_check_default: bool = True,
    ansible_core_package: str = "ansible-core",
    run_galaxy_importer: bool = True,
    galaxy_importer_package: str = "galaxy-importer",
    galaxy_importer_config_path: (
        str | os.PathLike | None
    ) = None,  # https://github.com/ansible/galaxy-importer#configuration
    galaxy_importer_always_show_logs: bool = False,
) -> None:
    """
    Add license-check session for license checks.
    """

    def compose_dependencies() -> list[str]:
        deps = [ansible_core_package]
        if run_galaxy_importer:
            deps.append(galaxy_importer_package)
        return deps

    def build_import_check(session: nox.Session) -> None:
        install(session, *compose_dependencies())

        tarball, _, __ = build_collection(session)

        if run_galaxy_importer and tarball:
            env = {}
            if galaxy_importer_config_path:
                env["GALAXY_IMPORTER_CONFIG"] = str(
                    Path(galaxy_importer_config_path).absolute()
                )
            assert tarball.parent is not None
            with session.chdir(tarball.parent), silence_run_verbosity():
                import_log = session.run(
                    "python",
                    "-m",
                    "galaxy_importer.main",
                    tarball.name,
                    env=env,
                    silent=True,
                )
            if import_log is not None:
                with ci_group("Run Galaxy importer") as (indent, is_collapsed):
                    if (
                        is_collapsed
                        or galaxy_importer_always_show_logs
                        or nox_has_verbosity()
                    ):
                        for line in import_log.splitlines():
                            print(f"{indent}{line}")
                        sys.stdout.flush()
                error_prefix = "ERROR:"
                errors = []
                for line in import_log.splitlines():
                    if line.startswith(error_prefix):
                        errors.append(line[len(error_prefix) :].strip())
                if errors:
                    messages = "\n".join(f"* {error}" for error in errors)
                    session.warn(
                        "Galaxy importer emitted the following non-fatal"
                        f" error{'' if len(errors) == 1 else 's'}:\n{messages}"
                    )

    build_import_check.__doc__ = compose_description(
        prefix={
            "one": "Run build and import checker:",
            "other": "Run build and import checkers:",
        },
        programs={
            "build-collection": True,
            "galaxy-importer": (
                "test whether Galaxy will import built collection"
                if run_galaxy_importer
                else False
            ),
        },
    )
    nox.session(
        name="build-import-check",
        default=make_build_import_check_default,
    )(build_import_check)


__all__ = [
    "add_build_import_check",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
General utilities.
"""

from __future__ import annotations

import typing as t
from dataclasses import dataclass


@dataclass(order=True, frozen=True)
class Version:
    """
    Models a two-part version (major and minor).
    """

    major: int
    minor: int

    @classmethod
    def parse(cls, version_string: str) -> Version:
        """
        Given a version string, parses it into a Version object.
        Other components than major and minor version are ignored.
        """
        try:
            major, minor = [int(v) for v in version_string.split(".")[:2]]
        except (AttributeError, ValueError) as exc:
            raise ValueError(
                f"Cannot parse {version_string!r} as version string."
            ) from exc
        return Version(major=major, minor=minor)

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}"

    def next_minor_version(self) -> Version:
        """
        Returns the next minor version.

        The major version stays the same, and the minor version is increased by 1.
        """
        return Version(self.major, self.minor + 1)

    def previous_minor_version(self) -> Version:
        """
        Returns the previous minor version.
        Raises a ValueError if there is none.

        The major version stays the same, and the minor version is decreased by 1.
        """
        if self.minor == 0:
            raise ValueError("No previous minor version exists")
        return Version(self.major, self.minor - 1)


def version_range(
    start: Version, end: Version, *, inclusive: bool
) -> t.Generator[Version]:
    """
    Enumerate all versions starting at ``start`` until ``end``.

    Whether ``end`` is included in the range depends on ``inclusive``.
    """
    if start.major != end.major:
        raise ValueError(
            f"Cannot list all versions from {start} to {end}"
            " since they have different major versions"
        )
    version = start
    while (version <= end) if inclusive else (version < end):
        yield version
        version = Version(major=version.major, minor=version.minor + 1)


__all__ = [
    "Version",
    "version_range",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Antsibull Nox Helper.
"""

from __future__ import annotations

from .config import (
    CONFIG_FILENAME,
    load_config_from_toml,
)
from .interpret_config import interpret_config
from .sessions.ansible_test import add_ansible_test_session

__version__ = "1.1.0"


def load_antsibull_nox_toml() -> None:
    """
    Load and interpret antsibull-nox.toml config file.
    """
    config = load_config_from_toml(CONFIG_FILENAME)
    interpret_config(config)


# pylint:disable=duplicate-code
__all__ = (
    "__version__",
    "add_ansible_test_session",
    "load_antsibull_nox_toml",
)



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Container engine related tools
"""

from __future__ import annotations

import functools
import os
import shutil
import subprocess
import typing as t

ContainerEngineType = t.Literal["docker", "podman"]
ContainerEnginePreferenceType = t.Literal[
    "docker", "podman", "auto", "auto-prefer-docker", "auto-prefer-podman"
]

ANTSIBULL_NOX_CONTAINER_ENGINE = "ANTSIBULL_NOX_CONTAINER_ENGINE"
VALID_CONTAINER_ENGINE_PREFERENCES: set[ContainerEnginePreferenceType] = {
    "docker",
    "podman",
    "auto",
    "auto-prefer-docker",
    "auto-prefer-podman",
}
DEFAULT_CONTAINER_ENGINE_PREFERENCE: ContainerEnginePreferenceType = "auto"


@functools.cache
def get_container_engine_preference() -> tuple[ContainerEnginePreferenceType, bool]:
    """
    Get the container engine preference.
    """
    container_engine = os.environ.get(ANTSIBULL_NOX_CONTAINER_ENGINE)
    if not container_engine:
        return DEFAULT_CONTAINER_ENGINE_PREFERENCE, False
    if container_engine not in VALID_CONTAINER_ENGINE_PREFERENCES:
        allowed_values = ", ".join(sorted(VALID_CONTAINER_ENGINE_PREFERENCES))
        raise ValueError(
            f"Invalid value for {ANTSIBULL_NOX_CONTAINER_ENGINE}: {container_engine!r}."
            f" Expected one of: {allowed_values}"
        )
    return t.cast(ContainerEnginePreferenceType, container_engine), True


def _get_version(exec_path: str) -> str | None:
    try:
        completed = subprocess.run(
            [exec_path, "-v"], capture_output=True, check=True, encoding="utf-8"
        )
        return completed.stdout.strip()
    except Exception:  # pylint: disable=broad-exception-caught
        return None


@functools.cache
def get_preferred_container_engine() -> ContainerEngineType:
    """
    Get the name of the preferred container engine.
    """
    preference = get_container_engine_preference()[0]
    if preference in ("podman", "docker"):
        return t.cast(ContainerEngineType, preference)
    docker_path = shutil.which("docker")
    podman_path = shutil.which("podman")
    if podman_path:
        version = _get_version(podman_path)
        if version is None:
            podman_path = None
    if docker_path:
        version = _get_version(docker_path)
        if version is None:
            docker_path = None
        elif "podman" in version and podman_path is not None:
            docker_path = None
    print(preference, docker_path, podman_path)
    if docker_path and podman_path:
        # Prefer one over the other by user's preference
        if preference == "auto-prefer-docker":
            podman_path = None
        if preference == "auto-prefer-podman":
            docker_path = None
    # Pick one
    if docker_path:
        return "docker"
    if podman_path:
        return "podman"
    raise ValueError("Could neither find 'docker' or 'podman' CLI on path!")


__all__ = (
    "get_container_engine_preference",
    "get_preferred_container_engine",
)



#!/usr/bin/env python

# Copyright (c) 2024, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt
# or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""Make sure all modules that should show up in the action group."""

from __future__ import annotations

import os
import re
import sys
import typing as t

import yaml

from antsibull_nox.data.antsibull_nox_data_util import setup
from antsibull_nox.sessions.extra_checks import ActionGroup


def compile_patterns(
    config: list[ActionGroup], errors: list[str]
) -> dict[str, re.Pattern] | None:
    patterns: dict[str, re.Pattern] = {}
    for action_group in config:
        if action_group.name in config:
            errors.append(
                f"noxfile.py: Action group {action_group.name!r} defined multiple times"
            )
            return None
        patterns[action_group.name] = re.compile(action_group.pattern)
    return patterns


def load_redirects(
    config: list[ActionGroup], errors: list[str], meta_runtime: str
) -> dict[str, list[str]]:
    # Load redirects
    try:
        with open(meta_runtime, "rb") as f:
            data = yaml.safe_load(f)
        action_groups = data.get("action_groups", {})
    except Exception as exc:
        errors.append(f"{meta_runtime}: cannot load action groups: {exc}")
        return {}

    if not isinstance(action_groups, dict):
        errors.append(f"{meta_runtime}: action_groups is not a dictionary")
        return {}
    if not all(
        isinstance(k, str) and isinstance(v, list) for k, v in action_groups.items()
    ):
        errors.append(
            f"{meta_runtime}: action_groups is not a dictionary mapping strings to list of strings"
        )
        return {}

    # Compare meta/runtime.yml content with config
    config_groups = {cfg.name for cfg in config}
    for action_group, elements in action_groups.items():
        if action_group not in config_groups:
            if len(elements) == 1 and isinstance(elements[0], dict):
                # Special case: if an action group is there with a single metadata entry,
                # we don't complain that it shouldn't be there.
                continue
            errors.append(
                f"{meta_runtime}: found unknown action group"
                f" {action_group!r}; likely noxfile needs updating"
            )
        else:
            action_groups[action_group] = [
                element for element in elements if isinstance(element, str)
            ]
    for action_group in config:
        if action_group.name not in action_groups:
            errors.append(
                f"{meta_runtime}: cannot find action group"
                f" {action_group.name!r}; likely noxfile needs updating"
            )

    return action_groups


def load_docs(path: str, errors: list[str]) -> dict[str, t.Any] | None:
    documentation = []
    in_docs = False
    with open(path, "r", encoding="utf-8") as f:
        for line in f:
            if line.startswith("DOCUMENTATION ="):
                in_docs = True
            elif line.startswith(("'''", '"""')) and in_docs:
                in_docs = False
            elif in_docs:
                documentation.append(line)
    if in_docs:
        errors.append(f"{path}: cannot find DOCUMENTATION end")
    if not documentation:
        errors.append(f"{path}: cannot find DOCUMENTATION")
        return None

    try:
        docs = yaml.safe_load("\n".join(documentation))
        if not isinstance(docs, dict):
            raise Exception("is not a top-level dictionary")
        return docs
    except Exception as exc:
        errors.append(f"{path}: cannot load DOCUMENTATION as YAML: {exc}")
        return None


def scan(config: list[ActionGroup], errors: list[str]) -> None:
    patterns = compile_patterns(config, errors)
    if patterns is None:
        return

    meta_runtime = "meta/runtime.yml"
    action_groups = load_redirects(config, errors, meta_runtime)

    modules_directory = "plugins/modules/"
    modules_suffix = ".py"

    for file in os.listdir(modules_directory):
        if not file.endswith(modules_suffix):
            continue
        module_name = file[: -len(modules_suffix)]

        for action_group in config:
            action_group_content = action_groups.get(action_group.name) or []
            path = os.path.join(modules_directory, file)

            if not patterns[action_group.name].match(module_name):
                if module_name in action_group_content:
                    errors.append(
                        f"{path}: module is in action group {action_group.name!r}"
                        " despite not matching its pattern as defined in noxfile"
                    )
                continue

            should_be_in_action_group = (
                module_name not in action_group.exclusions
                if action_group.exclusions
                else True
            )

            if should_be_in_action_group:
                if module_name not in action_group_content:
                    errors.append(
                        f"{meta_runtime}: module {module_name!r} is not part"
                        f" of {action_group.name!r} action group"
                    )
                else:
                    action_group_content.remove(module_name)

            docs = load_docs(path, errors)
            if docs is None:
                continue

            docs_fragments = docs.get("extends_documentation_fragment") or []
            is_in_action_group = action_group.doc_fragment in docs_fragments

            if should_be_in_action_group != is_in_action_group:
                if should_be_in_action_group:
                    errors.append(
                        f"{path}: module does not document itself as part of"
                        f" action group {action_group.name!r}, but it should;"
                        f" you need to add {action_group.doc_fragment} to"
                        f' "extends_documentation_fragment" in DOCUMENTATION'
                    )
                else:
                    errors.append(
                        f"{path}: module documents itself as part of"
                        f" action group {action_group.name!r}, but it should not be"
                    )

    for action_group in config:
        action_group_content = action_groups.get(action_group.name) or []
        for module_name in action_group_content:
            errors.append(
                f"{meta_runtime}: module {module_name} mentioned"
                f" in {action_group.name!r} action group does not exist"
                " or does not match pattern defined in noxfile"
            )


def main() -> int:
    """Main entry point."""
    paths, extra_data = setup()

    if not isinstance(extra_data.get("config"), list):
        raise ValueError("config is not a list")
    if not all(isinstance(cfg, dict) for cfg in extra_data["config"]):
        raise ValueError("config is not a list of dictionaries")
    config = [ActionGroup(**cfg) for cfg in extra_data["config"]]

    errors: list[str] = []
    scan(config, errors)

    for error in sorted(errors):
        print(error)
    return len(errors) > 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

# Copyright (c) Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt
# or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""Prevent unwanted files from being added to the source tree."""

from __future__ import annotations

import os
import sys

from antsibull_nox.data.antsibull_nox_data_util import (
    get_list_of_strings,
    setup,
)


def main() -> int:
    """Main entry point."""
    paths, extra_data = setup()

    skip_paths = set(get_list_of_strings(extra_data, "skip_paths", default=[]))

    skip_directories = tuple(
        get_list_of_strings(extra_data, "skip_directories", default=[])
    )

    errors: list[str] = []
    for path in paths:
        if path in skip_paths:
            continue

        if any(path.startswith(skip_directory) for skip_directory in skip_directories):
            continue

        if not os.path.isfile(path):
            continue

        try:
            with open(path, "r", encoding="utf-8") as f:
                for i, line in enumerate(f):
                    line = line.rstrip("\n\r")
                    if line.rstrip() != line:
                        errors.append(f"{path}:{i + 1}: found trailing whitespace")
        except UnicodeDecodeError:
            errors.append(f"{path}: cannot parse file as UTF-8")
        except Exception as e:
            errors.append(f"{path}: unexpected error: {e}")

    for error in sorted(errors):
        print(error)
    return len(errors) > 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2025 Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt
# or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""
This is a test module.
"""


DOCUMENTATION = r"""
---
module: test
short_description: Test module
author:
  - Felix Fontein (@felixfontein)
description:
  - Does nothing.
notes:
  - Does not support C(check_mode).
"""

from ansible.module_utils.basic import (  # noqa: E402, pylint: disable=wrong-import-position
    AnsibleModule,
)


def main():
    """
    The module's main function.
    """
    module = AnsibleModule(argument_spec={})
    module.exit_json(msg="Hello!")


if __name__ == "__main__":
    main()
